#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Strip Equivalent Nodes Graph Rewriter."""

from neural_compressor.common import logger
from neural_compressor.tensorflow.quantization.utils.utility import fix_ref_type_of_graph_def, strip_equivalent_nodes
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class StripEquivalentNodesOptimizer(GraphRewriterBase):
    """Remove the equivalent nodes which have the same inputs and attributes."""

    def __init__(self, model, output_node_names):
        """Initialization."""
        super().__init__(model)
        self.output_node_names = output_node_names

    @dump_elapsed_time("Pass StripEquivalentNodesOptimizer")
    def do_transformation(self):
        """Strip the equivalent nodes in the graph."""
        self.model = fix_ref_type_of_graph_def(self.model)
        iter_num = 0
        replaced_nodes_type = True
        all_replaced_nodes_type = {}
        while replaced_nodes_type:
            self.model, replaced_nodes_type = strip_equivalent_nodes(self.model, self.output_node_names)
            for k, v in replaced_nodes_type.items():
                all_replaced_nodes_type[k] = all_replaced_nodes_type.get(k, 0) + v
            iter_num += 1
            logger.debug(
                f"StripEquivalentNodes[Iter-{iter_num}]-Replaced equivalent node types are {replaced_nodes_type}"
            )
        logger.warning("All replaced equivalent node types are {}".format(all_replaced_nodes_type))
        return self.model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Freeze value with calibration Graph Rewriter."""

import re

import numpy as np
from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper

from ..graph_base import GraphRewriterBase


class FreezeValueTransformer(GraphRewriterBase):
    """Freeze Value with calibration."""

    def __init__(self, model, max_min_data, postfix, tensor_data=None, th=1, device="gpu", itex_mode=False):
        """Free Max/Min value into QuantizeV2 op.

        Args:
            model (graphdef): input model
            max_min_data (string list): the string context contains max/min values.
            postfix (string): the specified postfix to locate value.
            tensor_data(dict): key is the op name while the value is the max/min values
                                which calculated by KL.
            th (float, optional): The percentage of overall data.Defaults to 0.95.
            device (string, optional): The hardware device type, 'cpu' or 'gpu'.
            itex_mode: itex mode.
        """
        super().__init__(model)
        self.data = max_min_data
        if 0.0 < th <= 1.0:
            self.threshold = th
        else:
            self.logger.warning("The threshold value for clipping is invalid, " "Reset it to 0.95 by default.")
            self.threshold = 0.95
        self.postfix = postfix
        self.device = device
        self.itex_mode = itex_mode
        self.tensor_data = tensor_data
        self.cur_graph = GraphAnalyzer()
        self.cur_graph.graph = self.model

        self.graph_info = self.cur_graph.parse_graph()
        self.cur_graph.get_frame_info()

        self.quantizeV2_min_max = {}
        self.requant_min_max = {}
        self.scale_info = {}

    def _get_valid_log(self):
        """Get valid sampling log.

        :return: get the valid sampling log output.
        """
        output = []

        target_lines = [i.strip() for i in self.data if i.strip().find(";") != -1]
        for i in target_lines:
            semi_count = i.count(";")
            if semi_count == 2:
                output.append(i)
            elif semi_count % 2 != 0:
                self.logger.warning("Invalid line.")
            else:  # pragma: no cover
                loop_times = int(semi_count / 2)
                semi_index = [index for index, value in enumerate(i) if value == ";"]
                for index in range(loop_times - 1):
                    output.append(i[semi_index[index * 2] : semi_index[index * 2 + 2]])
                output.append(i[semi_index[loop_times * 2 - 2] :])
        return output

    def _parse_max_min_log(self):
        """Parse the max_ming log file.

        :return: get the node name and value mapping
        """
        print_suffix = "__print__"
        lines = self._get_valid_log()

        res = {}
        temp = {}
        pattern_def = r"{};{}\[\-?\d+\.?\d*e?-?\+?\d*\]".format(print_suffix, self.postfix)
        for i in lines:
            if not re.search(pattern_def, i):
                continue

            max_line_data = i.split(";")
            name = max_line_data[1][: -len(print_suffix)]
            value = max_line_data[-1].split("[")[-1].split("]")[0]
            if "eightbit" in name and name not in temp:
                temp[name] = []
            if "eightbit" in name:
                temp[name].append(float(value))
        for key in temp:
            target_index = int(len(temp[key]) * self.threshold)
            if target_index > len(temp[key]) - 1:
                target_index = len(temp[key]) - 1
            if self.postfix == "__min:":
                res[key] = sorted(temp[key], reverse=True)[target_index]
            else:
                res[key] = sorted(temp[key])[target_index]
        return res

    def _parse_requantization_ranges(self):
        """Parse the max_min log to get requantization values.

        :return: dict saved the result
        """
        res = {}

        print_suffix = "__print__"
        lines = self._get_valid_log()
        temp_min = {}
        temp_max = {}
        pattern_def = r"{};{}:\[\-?\d+\.?\d*e?-?\+?\d*\]".format(print_suffix, self.postfix)
        for i in lines:
            if not re.search(pattern_def, i):
                continue

            max_line_data = i.split(print_suffix + ";" + self.postfix)[-1]
            min_value = max_line_data.split("][")[0].split("[")[1]
            max_value = max_line_data.split("][")[1].split("]")[0]
            name = i.split(";")[1].strip()[: -len(print_suffix)]
            if name not in temp_min:
                temp_min[name] = []
            if name not in temp_max:
                temp_max[name] = []

            temp_min[name].append(float(min_value))
            temp_max[name].append(float(max_value))

        for key in temp_min:
            target_min_index = int(np.ceil(len(temp_min[key]) * (1 - self.threshold)))

            if key not in res:
                res[key] = []

            if target_min_index > len(temp_min[key]) - 1:
                target_min_index = len(temp_min[key]) - 1
            res[key].append(sorted(temp_min[key])[target_min_index])

        for key in temp_max:
            target_max_index = int(np.floor(len(temp_max[key]) * self.threshold)) - 1

            if target_max_index > len(temp_max[key]) - 1:
                target_max_index = len(temp_max[key]) - 1

            res[key].append(sorted(temp_max[key])[target_max_index])

        if self.tensor_data:  # pragma: no cover
            for k, v in self.tensor_data.items():
                if k in res:
                    self.logger.debug("Update node {} min to {}, max to {}.".format(k, v[2], v[3]))
                    res[k] = [v[2], v[3]]
        return res

    def generate_output_graph(self, max_name_value):
        """Generate transformed graph for freeze_max/freeze_min transformation.

        :param max_name_value: target values
        :return: transformed graph
        """
        for node_name, value in max_name_value.items():
            node_name = node_name.replace(":", "__port__").replace("^", "__hat__")
            if node_name not in self.graph_info:
                continue
            new_node = node_def_pb2.NodeDef()
            new_node.op = "Const"
            new_node_postfix = "/frozen_{}_only".format("".join([x for x in self.postfix if x.isalpha()]))
            new_node.name = node_name + new_node_postfix
            new_node.attr["dtype"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
            new_node.attr["value"].CopyFrom(
                attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(float(value), dtypes.float32, []))
            )
            output_node_name = self.graph_info[node_name].outputs[0]

            if (
                not self.itex_mode
                and node_name in self.cur_graph.parent_frame_details
                and self.cur_graph.parent_frame_details[node_name]
            ):  # pragma: no cover
                new_node_enter_node = Helper.create_node("Enter", new_node.name + "_enter", [new_node.name])
                Helper.set_attr_string(
                    new_node_enter_node,
                    "frame_name",
                    self.cur_graph.parent_frame_details[node_name].attr["frame_name"].s,
                )
                Helper.set_attr_dtype(new_node_enter_node, "T", dtypes.float32)
                Helper.set_attr_bool(new_node_enter_node, "is_constant", True)
                Helper.set_attr_int(
                    new_node_enter_node,
                    "parallel_iterations",
                    self.cur_graph.parent_frame_details[node_name].attr["parallel_iterations"].i,
                )

                self.cur_graph.add_node(new_node, None, [new_node_enter_node.name])
                # self.cur_graph.add_node(new_node_enter_node, new_node.name,
                #        [Helper.node_name_from_input(output_node_name)])
                self.cur_graph.replace_const_node(
                    new_node_enter_node, [Helper.node_name_from_input(output_node_name)], node_name
                )
                self.cur_graph.remove_node(node_name)
            else:
                self.cur_graph.replace_const_node(new_node, [Helper.node_name_from_input(output_node_name)], node_name)
                self.cur_graph.remove_node(node_name)

            self.quantizeV2_min_max[node_name] = tensor_util.MakeNdarray(new_node.attr["value"].tensor)
        self.scale_info[self.postfix[:-1]] = self.quantizeV2_min_max

        return GraphAnalyzer().dump_graph(), self.scale_info

    def generate_output_graph_ranges(self, max_name_value):
        """Generate transformed graph for freeze_max/freeze_min transformation.

        :param max_name_value: target values
        :return: transformed graph
        """
        for node_name, value in max_name_value.items():
            bn_node_name = node_name.replace("eightbit_requant_range", "eightbit_quantized_bn")
            in_node_name = node_name.replace("eightbit_requant_range", "eightbit_quantized_in")
            if not self.graph_info.get(bn_node_name) or not bn_node_name.endswith("_eightbit_quantized_bn"):
                bn_node_name = None
            if not self.graph_info.get(in_node_name) or not in_node_name.endswith("_eightbit_quantized_in"):
                in_node_name = None

            if self.itex_mode and "BatchNorm" in node_name:  # pragma: no cover
                bn_node_name = node_name[: -len("_eightbit_requant_range")]
                if bn_node_name not in self.graph_info:
                    bn_node_name = None
                else:
                    if "FusedBatchNorm" not in self.graph_info[bn_node_name].node.op:
                        bn_node_name = None

            if (
                node_name not in self.graph_info
                and bn_node_name not in self.graph_info
                and in_node_name not in self.graph_info
            ):
                continue

            min_node = node_def_pb2.NodeDef()
            min_node.op = "Const"
            min_node_postfix = "/frozen_min"
            if bn_node_name:
                min_node.name = bn_node_name + "/frozen_bn_output_min"
            elif in_node_name:
                min_node.name = in_node_name + "/frozen_in_output_min"
            else:
                min_node.name = node_name + min_node_postfix
            min_node.attr["dtype"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
            min_node.attr["value"].CopyFrom(
                attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(float(value[0]), dtypes.float32, []))
            )

            max_node = node_def_pb2.NodeDef()
            max_node.op = "Const"
            max_node_postfix = "/frozen_max"
            if bn_node_name:
                max_node.name = bn_node_name + "/frozen_bn_output_max"
            elif in_node_name:
                max_node.name = in_node_name + "/frozen_in_output_max"
            else:
                max_node.name = node_name + max_node_postfix
            max_node.attr["dtype"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
            max_node.attr["value"].CopyFrom(
                attr_value_pb2.AttrValue(tensor=tensor_util.make_tensor_proto(float(value[1]), dtypes.float32, []))
            )

            if bn_node_name:  # pragma: no cover
                if self.itex_mode:
                    self.cur_graph.replace_const_node(
                        min_node,
                        [Helper.node_name_from_input(bn_node_name + "_eightbit_quantize_bn")],
                        bn_node_name + "_eightbit_input7_output_min",
                    )
                    self.cur_graph.replace_const_node(
                        max_node,
                        [Helper.node_name_from_input(bn_node_name + "_eightbit_quantize_bn")],
                        bn_node_name + "_eightbit_input8_output_max",
                    )
                else:
                    self.cur_graph.replace_const_node(
                        min_node, [Helper.node_name_from_input(bn_node_name)], bn_node_name + "_input7_output_min"
                    )
                    self.cur_graph.replace_const_node(
                        max_node, [Helper.node_name_from_input(bn_node_name)], bn_node_name + "_input8_output_max"
                    )
            elif in_node_name:
                self.cur_graph.replace_const_node(
                    min_node, [Helper.node_name_from_input(in_node_name)], in_node_name + "_input7_output_min"
                )
                self.cur_graph.replace_const_node(
                    max_node, [Helper.node_name_from_input(in_node_name)], in_node_name + "_input8_output_max"
                )
            elif (
                not self.itex_mode
                and node_name in self.cur_graph.parent_frame_details
                and self.cur_graph.parent_frame_details[node_name]
            ):  # pragma: no cover
                output_node_name = self.graph_info[node_name].outputs[0]
                min_node_enter_node = Helper.create_node("Enter", min_node.name + "_enter", [min_node.name])
                Helper.set_attr_string(
                    min_node_enter_node,
                    "frame_name",
                    self.cur_graph.parent_frame_details[node_name].attr["frame_name"].s,
                )
                Helper.set_attr_dtype(min_node_enter_node, "T", dtypes.float32)
                Helper.set_attr_bool(min_node_enter_node, "is_constant", True)
                Helper.set_attr_int(
                    min_node_enter_node,
                    "parallel_iterations",
                    self.cur_graph.parent_frame_details[node_name].attr["parallel_iterations"].i,
                )

                self.cur_graph.add_node(min_node, None, [min_node_enter_node.name])
                # self.cur_graph.add_node(min_node_enter_node, min_node.name,
                #            [Helper.node_name_from_input(output_node_name)])
                self.cur_graph.replace_const_node(
                    min_node_enter_node, [Helper.node_name_from_input(output_node_name)], node_name + ":0"
                )

                max_node_enter_node = Helper.create_node("Enter", max_node.name + "_enter", [max_node.name])
                Helper.set_attr_string(
                    max_node_enter_node,
                    "frame_name",
                    self.cur_graph.parent_frame_details[node_name].attr["frame_name"].s,
                )
                Helper.set_attr_dtype(max_node_enter_node, "T", dtypes.float32)
                Helper.set_attr_bool(max_node_enter_node, "is_constant", True)
                Helper.set_attr_int(
                    max_node_enter_node,
                    "parallel_iterations",
                    self.cur_graph.parent_frame_details[node_name].attr["parallel_iterations"].i,
                )

                self.cur_graph.add_node(max_node, None, [max_node_enter_node.name])
                # self.cur_graph.add_node(max_node_enter_node, max_node.name,
                #              [Helper.node_name_from_input(output_node_name)])
                self.cur_graph.replace_const_node(
                    max_node_enter_node, [Helper.node_name_from_input(output_node_name)], node_name + ":1"
                )

                self.cur_graph.remove_node(node_name)
            else:
                output_node_name = self.graph_info[node_name].outputs[0]
                self.cur_graph.replace_const_node(
                    min_node, [Helper.node_name_from_input(output_node_name)], node_name + ":0"
                )
                self.cur_graph.replace_const_node(
                    max_node, [Helper.node_name_from_input(output_node_name)], node_name + ":1"
                )
                self.cur_graph.remove_node(node_name)

                self.requant_min_max[node_name] = [
                    tensor_util.MakeNdarray(min_node.attr["value"].tensor),
                    tensor_util.MakeNdarray(max_node.attr["value"].tensor),
                ]

        self.scale_info[self.postfix] = self.requant_min_max
        # {
        #    "__requant_min_max":
        #                {requant_range_name:[min,max],}
        #    "__min":
        #                {node_name_min:[min]}
        #    "__max":
        #                {node_name_max:[max]}
        # }
        return GraphAnalyzer().dump_graph(), self.scale_info

    def do_transformation(self):
        """Apply the transformation of freeze value."""
        if self.postfix == "__requant_min_max":
            range_data = self._parse_requantization_ranges()

            return self.generate_output_graph_ranges(range_data)

        max_name_value = self._parse_max_min_log()
        return self.generate_output_graph(max_name_value)



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The utility functions and classes for Tensorflow."""

import importlib
import logging
import os
import pickle
import subprocess
import sys
import time
from functools import reduce
from typing import Callable, Dict

import cpuinfo
import numpy as np
import psutil
from pkg_resources import parse_version

from neural_compressor.common.utils import Statistics, logger

# Dictionary to store a mapping between algorithm names and corresponding algo implementation(function)
algos_mapping: Dict[str, Callable] = {}


def version1_lt_version2(version1, version2):
    """Check whether version1 is less than version2."""
    return parse_version(version1) < parse_version(version2)


def version1_gt_version2(version1, version2):
    """Check whether version1 is greater than version2."""
    return parse_version(version1) > parse_version(version2)


def version1_eq_version2(version1, version2):
    """Check whether version1 is equal to version2."""
    return parse_version(version1) == parse_version(version2)


def version1_gte_version2(version1, version2):
    """Check whether version1 is greater than version2 or is equal to it."""
    return parse_version(version1) > parse_version(version2) or parse_version(version1) == parse_version(version2)


def version1_lte_version2(version1, version2):
    """Check whether version1 is less than version2 or is equal to it."""
    return parse_version(version1) < parse_version(version2) or parse_version(version1) == parse_version(version2)


def register_algo(name):
    """Decorator function to register algorithms in the algos_mapping dictionary.

    Usage example:
        @register_algo(name=example_algo)
        def example_algo(model: tf.keras.Model, quant_config: StaticQuantConfig) -> tf.keras.Model:
            ...
    Args:
        name (str): The name under which the algorithm function will be registered.

    Returns:
        decorator: The decorator function to be used with algorithm functions.
    """

    def decorator(algo_func):
        algos_mapping[name] = algo_func
        return algo_func

    return decorator


def deep_get(dictionary, keys, default=None):
    """Get the dot key's item in nested dict.

    Usage example:
       person = {'person':{'name':{'first':'John'}}}
       deep_get(person, "person.name.first") will output 'John'.
    Args:
        dictionary (dict): The dict object to get keys
        keys (dict): The deep keys
        default (object): The return item if key not exists

    Returns:
        item: the item of the deep dot keys
    """
    return reduce(lambda d, key: d.get(key, default) if isinstance(d, dict) else default, keys.split("."), dictionary)


def itex_installed():
    """Check if the Intel® Extension for TensorFlow has been installed."""
    try:
        import intel_extension_for_tensorflow

        return True
    except:
        return False


def dump_elapsed_time(customized_msg=""):
    """Get the elapsed time for decorated functions.

    Args:
        customized_msg (string, optional): The parameter passed to decorator. Defaults to None.
    """

    def f(func):
        def fi(*args, **kwargs):
            start = time.time()
            res = func(*args, **kwargs)
            end = time.time()
            logging.getLogger("neural_compressor").info(
                "%s elapsed time: %s ms"
                % (customized_msg if customized_msg else func.__qualname__, round((end - start) * 1000, 2))
            )
            return res

        return fi

    return f


def combine_histogram(old_hist, arr):
    """Collect layer histogram for arr and combine it with old histogram."""
    new_max = np.max(arr)
    new_min = np.min(arr)
    new_th = max(abs(new_min), abs(new_max))
    (old_hist, old_hist_edges, old_min, old_max, old_th) = old_hist
    if new_th <= old_th:
        hist, _ = np.histogram(arr, bins=len(old_hist), range=(-old_th, old_th))
        return (old_hist + hist, old_hist_edges, min(old_min, new_min), max(old_max, new_max), old_th)
    else:
        old_num_bins = len(old_hist)
        old_step = 2 * old_th / old_num_bins
        half_increased_bins = int((new_th - old_th) // old_step + 1)
        new_num_bins = half_increased_bins * 2 + old_num_bins
        new_th = half_increased_bins * old_step + old_th
        hist, hist_edges = np.histogram(arr, bins=new_num_bins, range=(-new_th, new_th))
        hist[half_increased_bins : new_num_bins - half_increased_bins] += old_hist
        return (hist, hist_edges, min(old_min, new_min), max(old_max, new_max), new_th)


def get_all_fp32_data(data):
    """Get all the fp32 data."""
    return [float(i) for i in data.replace("[", " ").replace("]", " ").split(" ") if i.strip() and len(i) < 32]


def get_tensor_histogram(tensor_data, bins=2048):
    """Get the histogram of the tensor data."""
    max_val = np.max(tensor_data)
    min_val = np.min(tensor_data)
    th = max(abs(min_val), abs(max_val))
    hist, hist_edges = np.histogram(tensor_data, bins=2048, range=(-th, th))
    return (hist, hist_edges, min_val, max_val, th)


def singleton(cls):
    """Not displayed in API Docs.

    Singleton decorator.
    """
    instances = {}

    def _singleton(*args, **kw):
        """Create a singleton object."""
        if cls not in instances:
            instances[cls] = cls(*args, **kw)
        return instances[cls]

    return _singleton


def disable_random(seed=1):
    """A Decorator to disable tf random seed."""
    import tensorflow as tf

    def decorator(func):
        def wrapper(*args, **kw):
            tf.compat.v1.disable_eager_execution()
            tf.compat.v1.reset_default_graph()
            tf.compat.v1.set_random_seed(seed)
            return func(*args, **kw)

        return wrapper

    return decorator


def valid_keras_format(model):
    """Check if the input model is Sequential or Functional model."""
    import keras

    if isinstance(model, keras.src.models.Sequential) or isinstance(model, keras.src.models.Functional):
        return True

    return False


@singleton
class CpuInfo(object):
    """Get CPU Info."""

    def __init__(self):
        """Get whether the cpu numerical format is bf16, the number of sockets, cores and cores per socket."""
        self._bf16 = False
        self._vnni = False
        info = cpuinfo.get_cpu_info()
        if "arch" in info and "X86" in info["arch"]:
            cpuid = cpuinfo.CPUID()
            max_extension_support = cpuid.get_max_extension_support()
            if max_extension_support >= 7:
                ecx = cpuid._run_asm(
                    b"\x31\xC9",  # xor ecx, ecx
                    b"\xB8\x07\x00\x00\x00" b"\x0f\xa2" b"\x89\xC8" b"\xC3",  # mov eax, 7  # cpuid  # mov ax, cx  # ret
                )
                self._vnni = bool(ecx & (1 << 11))
                eax = cpuid._run_asm(
                    b"\xB9\x01\x00\x00\x00",  # mov ecx, 1
                    b"\xB8\x07\x00\x00\x00" b"\x0f\xa2" b"\xC3",  # mov eax, 7  # cpuid  # ret
                )
                self._bf16 = bool(eax & (1 << 5))
        if "arch" in info and "ARM" in info["arch"]:  # pragma: no cover
            self._sockets = 1
        else:
            self._sockets = self.get_number_of_sockets()
        self._cores = psutil.cpu_count(logical=False)
        self._cores_per_socket = int(self._cores / self._sockets)

    @property
    def bf16(self):
        """Get whether it is bf16."""
        return self._bf16

    @property
    def vnni(self):
        """Get whether it is vnni."""
        return self._vnni

    @property
    def cores_per_socket(self):
        """Get the cores per socket."""
        return self._cores_per_socket

    def get_number_of_sockets(self) -> int:
        """Get number of sockets in platform."""
        cmd = "cat /proc/cpuinfo | grep 'physical id' | sort -u | wc -l"
        if psutil.WINDOWS:
            cmd = r'wmic cpu get DeviceID | C:\Windows\System32\find.exe /C "CPU"'

        with subprocess.Popen(
            args=cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=False,
        ) as proc:
            proc.wait()
            if proc.stdout:
                for line in proc.stdout:
                    return int(line.decode("utf-8", errors="ignore").strip())
        return 0


class CaptureOutputToFile(object):
    """Not displayed in API Docs.

    Capture the output to file.
    """

    def __init__(self, tmp_file_path, stream=sys.stderr):
        """Open a temporary file."""
        self.orig_stream_fileno = stream.fileno()
        self.tmp_file = open(tmp_file_path, "w")

    def __enter__(self):
        """Duplicate the file descriptor to the stream."""
        self.orig_stream_dup = os.dup(self.orig_stream_fileno)
        os.dup2(self.tmp_file.fileno(), self.orig_stream_fileno)

    def __exit__(self, type, value, traceback):
        """Duplicate the stream descriptor to the file."""
        os.close(self.orig_stream_fileno)
        os.dup2(self.orig_stream_dup, self.orig_stream_fileno)
        os.close(self.orig_stream_dup)
        self.tmp_file.close()


@singleton
class TFSlimNetsFactory(object):  # pragma: no cover
    """TF-Slim nets factory."""

    def __init__(self):
        """Initialize a TFSlimNetsFactory."""
        # tf_slim only support specific models by default
        self.default_slim_models = [
            "alexnet_v2",
            "overfeat",
            "vgg_a",
            "vgg_16",
            "vgg_19",
            "inception_v1",
            "inception_v2",
            "inception_v3",
            "resnet_v1_50",
            "resnet_v1_101",
            "resnet_v1_152",
            "resnet_v1_200",
            "resnet_v2_50",
            "resnet_v2_101",
            "resnet_v2_152",
            "resnet_v2_200",
        ]

        from tf_slim.nets import alexnet, inception, overfeat, resnet_v1, resnet_v2, vgg

        self.networks_map = {
            "alexnet_v2": {
                "model": alexnet.alexnet_v2,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": alexnet.alexnet_v2_arg_scope,
            },
            "overfeat": {
                "model": overfeat.overfeat,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": overfeat.overfeat_arg_scope,
            },
            "vgg_a": {
                "model": vgg.vgg_a,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1000,
                "arg_scope": vgg.vgg_arg_scope,
            },
            "vgg_16": {
                "model": vgg.vgg_16,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1000,
                "arg_scope": vgg.vgg_arg_scope,
            },
            "vgg_19": {
                "model": vgg.vgg_19,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1000,
                "arg_scope": vgg.vgg_arg_scope,
            },
            "inception_v1": {
                "model": inception.inception_v1,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": inception.inception_v1_arg_scope,
            },
            "inception_v2": {
                "model": inception.inception_v2,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": inception.inception_v2_arg_scope,
            },
            "inception_v3": {
                "model": inception.inception_v3,
                "input_shape": [None, 299, 299, 3],
                "num_classes": 1001,
                "arg_scope": inception.inception_v3_arg_scope,
            },
            "resnet_v1_50": {
                "model": resnet_v1.resnet_v1_50,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1000,
                "arg_scope": resnet_v1.resnet_arg_scope,
            },
            "resnet_v1_101": {
                "model": resnet_v1.resnet_v1_101,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1000,
                "arg_scope": resnet_v1.resnet_arg_scope,
            },
            "resnet_v1_152": {
                "model": resnet_v1.resnet_v1_152,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1000,
                "arg_scope": resnet_v1.resnet_arg_scope,
            },
            "resnet_v1_200": {
                "model": resnet_v1.resnet_v1_200,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1000,
                "arg_scope": resnet_v1.resnet_arg_scope,
            },
            "resnet_v2_50": {
                "model": resnet_v2.resnet_v2_50,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": resnet_v2.resnet_arg_scope,
            },
            "resnet_v2_101": {
                "model": resnet_v2.resnet_v2_101,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": resnet_v2.resnet_arg_scope,
            },
            "resnet_v2_152": {
                "model": resnet_v2.resnet_v2_152,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": resnet_v2.resnet_arg_scope,
            },
            "resnet_v2_200": {
                "model": resnet_v2.resnet_v2_200,
                "input_shape": [None, 224, 224, 3],
                "num_classes": 1001,
                "arg_scope": resnet_v2.resnet_arg_scope,
            },
        }

    def register(self, name, model_func, input_shape, arg_scope, **kwargs):
        """Register a model to TFSlimNetsFactory.

        Args:
            name (str): name of a model.
            model_func (_type_): model that built from slim.
            input_shape (_type_): input tensor shape.
            arg_scope (_type_): slim arg scope that needed.
        """
        net_info = {"model": model_func, "input_shape": input_shape, "arg_scope": arg_scope}
        net = {name: {**net_info, **kwargs}}
        self.networks_map.update(net)
        self.default_slim_models.append(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert Add OP to BiasAdd OP Graph Rewriter."""

import numpy as np
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_rewriter.graph_base import GraphRewriterBase
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import SPR_BASE_VERSIONS, dump_elapsed_time


class ConvertAddToBiasAddOptimizer(GraphRewriterBase):
    """Convert MatMul/Conv2D + Add(AddV2) to MatMul + BiasAdd."""

    @dump_elapsed_time("Pass ConvertAddToBiasAddOptimizer")
    def do_transformation(self):
        """Execute conversion Add to BiasAdd."""
        g = GraphAnalyzer()
        g.graph = self.model
        graph_info = g.parse_graph()

        import tensorflow as tf

        if tf.version.VERSION not in SPR_BASE_VERSIONS:
            target_nodes = g.query_fusion_pattern_nodes([["MatMul", "Conv2D"], ["Add", "AddV2"]])
        else:
            target_nodes = g.query_fusion_pattern_nodes([["MatMul"], ["Add", "AddV2"]])
        for i in target_nodes:
            successor_node_names = graph_info[i[1]].outputs
            matmul_input_name = graph_info[i[0]].node.input[0]
            matmul_input_node = graph_info[Helper.node_name_from_input(matmul_input_name)].node
            # Fixme below two lines was added due to MatMul kernel limitation for matmul input type
            # should be quint8.
            if matmul_input_node.op == "Const":
                continue
            add_second_input_name = graph_info[i[1]].node.input[1]
            add_second_const_node = graph_info[add_second_input_name].node
            if add_second_const_node.op != "Const":
                continue
            bias_tensor = tensor_util.MakeNdarray(add_second_const_node.attr["value"].tensor)

            if bias_tensor.ndim > 2:
                continue

            new_bias_tensor = np.ravel(bias_tensor)

            g.remove_node(i[1])

            bias_node_name = i[1]
            bias_const_node_name = add_second_const_node.name + "_flattern"

            bias_const_node = Helper.create_constant_node(bias_const_node_name, new_bias_tensor, dtypes.float32)

            bias_node = Helper.create_node("BiasAdd", bias_node_name, [i[0], bias_const_node_name])
            Helper.set_attr_dtype(bias_node, "T", dtypes.float32)

            g.add_node(bias_const_node, None, [bias_node_name])
            g.replace_single_node(bias_node, [i[0]], i[1], successor_node_names, i[1])

        return g.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""QuantizeGraph Base Class."""

import logging
from collections import namedtuple

import tensorflow as tf
from tensorflow.core.framework import graph_pb2
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper
from neural_compressor.tensorflow.utils import version1_eq_version2, version1_gt_version2, version1_lt_version2


class QuantizeGraphBase:
    """This is the base class for quantize graph."""

    def __init__(self, output_node_names):
        """Initilizaiton."""
        self.output_node_names = output_node_names
        self.transformers = {}

    def register_transformer(self, node_name, entry):
        """Register transformers API."""
        if node_name not in self.transformers:
            self.transformers[node_name] = {}

        self.transformers[node_name] = entry

    def do_transform(self):
        """This is the virtual interface need to be implemented by derived class.

        :return:
        """
        pass

    def remove_dead_nodes(self, input_graph, output_names):
        """Removes nodes that are no longer needed for inference from the graph."""
        return tf.compat.v1.graph_util.extract_sub_graph(input_graph, output_names)


class QuantizeNodeBase:
    """This is the base class for nodes fusion.

    Arguments:
        object {[type]} -- [description]
    """

    node_details = namedtuple("node_details", ["node", "output"])

    def __init__(self, **kwargs):
        """Initilizaiton."""
        self.logger = logging.getLogger("neural_compressor")
        input_graph = kwargs["input_graph"]

        assert isinstance(input_graph, graph_pb2.GraphDef)

        self.input_graph = input_graph

        self._parse_graph()
        self.output_node_maps = {}
        self.output_graph = graph_pb2.GraphDef()
        self.quantized_node_dict = {}
        self.patterns = kwargs["patterns"]
        self.remove_redundant_quant_flag = kwargs["remove_redundant_quant_flag"]
        self.fake_quant = kwargs["fake_quant"] if "fake_quant" in kwargs else False
        self.frame_info = kwargs["frame_info"] if "frame_info" in kwargs else None
        self.per_channel, self.is_asymmetric = kwargs["op_wise_cfg"][0], kwargs["op_wise_cfg"][2]
        self.op_wise_config_name_list = kwargs["op_wise_config_name_list"]
        self.weight_bit = kwargs["op_wise_cfg"][3]
        self.start_node_name = kwargs["start_node_name"]
        self.device = kwargs["device"]
        self.new_api = kwargs["new_api"]
        self.performance_only = kwargs["performance_only"]
        self.itex_mode = kwargs["itex_mode"]
        self.enable_s8 = bool(
            version1_gt_version2(tf.version.VERSION, "2.1.0")
            or version1_eq_version2(tf.version.VERSION, "2.1.0")
            or tf.version.VERSION.find("1.15.0-up") != -1
        )

    def apply_the_transform(self):
        """This is the virtual interface to be implemented by derived class.

        :return: transformed graphdef
        """
        pass

    def get_longest_fuse(self):
        """This is the virtual interface to be implemented by derived class."""
        pass

    def _insert_dummy_biasadd(self, match_node_name, matched_node):  # pragma: no cover
        """Insert dummy BiasAdd op."""
        target_node_name = matched_node.node.name
        matmul_a_node_name = helper.node_name_from_input(matched_node.node.input[0])
        matmul_a_node = self.node_name_mapping[matmul_a_node_name].node
        matmul_b_node_name = helper.node_name_from_input(matched_node.node.input[1])
        matmul_b_node = self.node_name_mapping[matmul_b_node_name].node

        if matmul_a_node.op == "Const" and matmul_b_node.op != "Const":
            pass
        else:
            from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer

            g = GraphAnalyzer()
            g.graph = self.input_graph
            graph_info = g.parse_graph()
            next_node_names = graph_info[matched_node.node.name].outputs
            bias_node_name = target_node_name + "_dummy_biasadd"
            bias_const_node_name = target_node_name + "_fake_const"

            if matched_node.node.op == "MatMul":
                t_b_index = 0 if matched_node.node.attr["transpose_b"].b else 1
            if (
                matched_node.node.op in ("Conv2D" or "DepthwiseConv2dNative")
                and matched_node.node.attr["data_format"].s == b"NHWC"
            ):
                t_b_index = 3
            elif (
                matched_node.node.op in ("Conv2D" or "DepthwiseConv2dNative")
                and matched_node.node.op.attr["data_format"].s == b"NCHW"
            ):
                t_b_index = 1
            elif matched_node.node.op == "Conv3D" and matched_node.node.attr["data_format"].s == b"NDHWC":
                t_b_index = 4
            elif matched_node.node.op == "Conv3D" and matched_node.node.attr["data_format"].s == b"NCDHW":
                t_b_index = 1
            bias_add_length = matmul_b_node.attr["value"].tensor.tensor_shape.dim[t_b_index].size

            bias_add_content = [0.0] * bias_add_length

            bias_const_node = helper.create_constant_node(
                bias_const_node_name, bias_add_content, dtypes.float32, shape=[bias_add_length]
            )
            bias_node = helper.create_node("BiasAdd", bias_node_name, [target_node_name, bias_const_node_name])
            helper.set_attr_dtype(bias_node, "T", dtypes.float32)

            g.add_node(bias_node, target_node_name, next_node_names)
            g.add_node(bias_const_node, None, [bias_node_name])
            self.input_graph = g.dump_graph()
            self._parse_graph(self.input_graph)
            new_match_node_name = match_node_name[:1] + [bias_node_name] + match_node_name[1:]
            new_match_node_name = match_node_name[:1] + [bias_node_name] + match_node_name[1:]

            return new_match_node_name

    def _is_match(self, patterns):
        """Detect the rule matched nodes collections.

        Returns:
            [List] -- [the matched rule]
            [String] -- [the list contains the matched node name]
        """
        matched_node_name = []

        for k, v in enumerate(self.op_list):
            if v in set(fusion[0] for fusion in patterns):
                cur_node = self.node_name_mapping[list(self.node_name_mapping.keys())[k]].node
                if cur_node.name != self.start_node_name:
                    continue

                if ((v in ("Conv2D", "DepthwiseConv2dNative") and not self.enable_s8)) and not self._find_relu_node(
                    cur_node
                ):
                    continue

                for sub_rule in patterns:
                    if v != sub_rule[0]:
                        continue

                    sub_rule_len = len(sub_rule)
                    check_hardswish = True if sub_rule_len > 4 else False
                    self.logger.debug("Try to apply rule: {}".format(sub_rule))

                    cur_node_name = list(self.node_name_mapping.keys())[k]

                    matched_node_name.clear()

                    matched_node_name.append(cur_node_name)
                    count = 0
                    while sub_rule_len > 1:
                        if not self.node_name_mapping[cur_node_name].output:
                            self.logger.debug("Fail to match {}".format(sub_rule))
                            break

                        next_node_name = self.node_name_mapping[cur_node_name].output[0]

                        is_shared_output = True if len(self.node_name_mapping[cur_node_name].output) > 1 else False

                        add_op_quantizable = True
                        is_hardswish = False
                        if is_shared_output:  # pragma: no cover
                            if next_node_name.find("hard_swish") != -1:
                                self.logger.debug("Find Hard Swish pattern ......")
                                is_hardswish = True
                                count = count + 1
                                if next_node_name.find("add") == -1:
                                    next_node_name = self.node_name_mapping[cur_node_name].output[1]
                            else:
                                add_op_quantizable = False
                        next_node_op = self.node_name_mapping[next_node_name].node.op

                        if next_node_op in ("Add", "AddV2", "AddN"):
                            next_node = self.node_name_mapping[next_node_name].node
                            next_node_inputs = list(next_node.input)
                            cur_node_index = next_node_inputs.index(cur_node_name)

                            for index, input_name in enumerate(next_node_inputs):
                                node_type = self.node_name_mapping[helper.node_name_from_input(input_name)].node.op
                                if input_name != cur_node_name and index < cur_node_index and node_type != "Dequantize":
                                    add_op_quantizable = False
                                    break

                        if add_op_quantizable and next_node_op == sub_rule[1 - sub_rule_len]:
                            if not is_shared_output:
                                matched_node_name.append(next_node_name)
                                sub_rule_len -= 1
                                cur_node_name = next_node_name
                            elif is_hardswish and self.new_api:  # pragma: no cover
                                matched_node_name.append(next_node_name)
                                sub_rule_len -= 1
                                cur_node_name = next_node_name
                            else:
                                matched_node_name.clear()
                                self.logger.debug("Fail to match {}.".format(sub_rule))
                                break
                        else:
                            matched_node_name.clear()
                            self.logger.debug("Fail to match {}.".format(sub_rule))
                            break

                    if sub_rule_len == 1:
                        if (
                            check_hardswish
                            and sub_rule[-1] == "Mul"
                            and sub_rule[-2] == "Mul"
                            and sub_rule[-3] == "Relu6"
                            and sub_rule[-4] == "Add"
                            and count != 1
                        ):
                            matched_node_name.clear()
                            self.logger.debug("Fail to match {}.".format(sub_rule))
                            break
                        self.logger.debug("Match {} on nodes {}.".format(sub_rule, matched_node_name))
                        return sub_rule, matched_node_name

        return None, None

    def _need_to_check(self, node_type):
        """Check op list."""
        op_list = (
            "ConcatV2",
            "Conv2D",
            "Conv3D",
            "DepthwiseConv2D",
            "QuantizeV2",
            "DepthwiseConv2dNative",
            "MaxPool",
            "MaxPool3D",
            "FusedBatchNormV3",
            "Requantize",
            "RequantizePerChannel",
            "AvgPool",
            "Pad",
            "CropAndResize",
            "Dequantize",
            "Mean",
            "MatMul",
            "BatchMatMulV2",
            "FakeQuantWithMinMaxVars",
            "_MklFusedInstanceNorm",
        )
        return any([node_type.find(i) != -1 for i in op_list])

    def _find_relu_node(self, node):
        """Find relu node algorithm to identify the positive input."""
        # if node.op.find("HardSwish") != -1:
        #    return False
        if (
            node.op in ("Relu", "Relu6")
            or (
                node.op.find("AndRelu") != -1
                and ("alpha" not in node.attr or ("alpha" in node.attr and node.attr["alpha"].f == 0))
            )
        ) and (
            node.op != "Relu"
            or not self.new_api
            or not self.performance_only
            or self.node_name_mapping[helper.node_name_from_input(node.input[0])].node.op.find("FusedBatchNorm") == -1
            or self.node_name_mapping[helper.node_name_from_input(node.input[0])].node.attr["is_training"].b
            or len(self.node_name_mapping[helper.node_name_from_input(node.input[0])].output) > 1
        ):
            return True
        elif "T" in node.attr and node.attr["T"].type in (dtypes.quint8, dtypes.uint8):
            return True
        elif (
            node.op.find("QuantizedConv") != -1
            or node.op.find("QuantizedDepthwiseConv") != -1
            or node.op.find("QuantizedMatMul") != -1
        ) and (node.op.find("Relu") == -1 or ("alpha" in node.attr and node.attr["alpha"].f > 0)):
            return False
        elif self._need_to_check(node.op):
            input_node = self.node_name_mapping[helper.node_name_from_input(node.input[0])]
            return self._find_relu_node(input_node.node)
        else:
            return False

    def _reset_output_node_maps(self):
        """Reset output node maps."""
        self.output_node_maps = {}

    def _get_op_list(self):
        """Get the op list."""
        self.op_list = []
        for _, v in enumerate(self.node_name_mapping):
            self.op_list.append(self.node_name_mapping[v].node.op)

    def _get_node_input(self, node_name):
        """Return control_input name, non-control_input node name."""
        return [i for i in self.node_name_mapping[node_name].node.input if i[0] == "^"], [
            i for i in self.node_name_mapping[node_name].node.input if i[0] != "^"
        ]

    def _intel_cpu_add_dequantize_result_node(
        self, quantized_output_name, original_node_name, dtype=dtypes.quint8, min_tensor_index=1, performance_only=False
    ):
        """Add Dequantize node after the quantized node."""
        min_max_inputs = [
            "%s:%s" % (quantized_output_name, min_tensor_index),
            "%s:%s" % (quantized_output_name, min_tensor_index + 1),
        ]
        dequantize_name = original_node_name

        dequantize_node = helper.create_node(
            "Dequantize", dequantize_name, [quantized_output_name, min_max_inputs[0], min_max_inputs[1]]
        )
        helper.set_attr_dtype(dequantize_node, "T", dtype)
        if performance_only:
            helper.set_attr_string(dequantize_node, "mode", b"SCALED")
        else:
            helper.set_attr_string(dequantize_node, "mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED")
        self.add_output_graph_node(dequantize_node)

    def eightbitize_single_input_tensor_node(self, original_node, add_op_function):
        """Quantize the single input tensor node."""
        quantized_op_name = original_node.name + "_eightbit_quantized"
        quantized_op_type = "Quantized" + original_node.op
        if version1_gt_version2(tf.version.VERSION, "2.7.0") and original_node.op == "MaxPool3D":
            quantized_op_type = "_Quantized" + original_node.op
        all_input_names = self._add_eightbit_prologue_nodes(original_node.name)
        quantized_op_node = helper.create_node(quantized_op_type, quantized_op_name, all_input_names)
        add_op_function(original_node, quantized_op_node)
        self.add_output_graph_node(quantized_op_node)
        deq_type = dtypes.quint8 if self._find_relu_node(original_node) else dtypes.qint8
        self._intel_cpu_add_dequantize_result_node(quantized_op_name, original_node.name, dtype=deq_type)

    def _add_eightbit_prologue_nodes(self, original_node):
        """Add quantized prologue nodes."""
        namespace_prefix = original_node + "_eightbit"
        reshape_dims_name, reduction_dims_name = self._add_common_quantization_nodes(
            namespace_prefix, helper.node_name_from_input(self.node_name_mapping[original_node].node.input[0])
        )
        input_names = []
        min_max_names = []
        for each_input_name in self.node_name_mapping[original_node].node.input[:1]:
            if each_input_name[0] == "^":
                continue

            if self.node_name_mapping[original_node].node.op == "MatMul":
                # mkl ops _MklQuantizedMatMulWithBiasAndRelu|AndRequantize
                # requires the T1 data type as quint8
                dtype = dtypes.quint8
            else:
                input_node_name = helper.node_name_from_input(each_input_name)
                if input_node_name in self.output_node_maps:
                    if self.output_node_maps[input_node_name].op == "Dequantize":
                        dtype = dtypes.DType(self.output_node_maps[input_node_name].attr["T"].type)
                    elif self._find_relu_node(self.node_name_mapping[original_node].node):
                        dtype = dtypes.quint8
                    else:
                        dtype = dtypes.qint8
                else:
                    dtype = (
                        dtypes.quint8
                        if self._find_relu_node(self.node_name_mapping[original_node].node)
                        else dtypes.qint8
                    )
                    if "FusedBatchNorm" in self.node_name_mapping[original_node].node.op:
                        dtype = dtypes.qint8
            quantize_input_name, min_input_name, max_input_name = self._eightbitize_input_to_node(
                namespace_prefix, each_input_name, reshape_dims_name, reduction_dims_name, dtype=dtype
            )
            input_names.append(quantize_input_name)
            min_max_names.append(min_input_name)
            min_max_names.append(max_input_name)
        all_input_names = []
        all_input_names.extend(input_names)
        if min_max_names:
            all_input_names.extend(min_max_names)

        for original_input_name in self.node_name_mapping[original_node].node.input:
            if original_input_name[0] == "^":
                all_input_names.append(original_input_name)
        return all_input_names

    def _add_eightbit_prologue_nodes_for_enter(self, original_node, enter_node=None):
        """Add quantized prologue nodes for control edge case."""
        namespace_prefix = original_node + "_eightbit"
        reshape_dims_name, reduction_dims_name = self._add_common_quantization_nodes(
            namespace_prefix,
            helper.node_name_from_input(self.node_name_mapping[original_node].node.input[0]),
            enter_node=enter_node,
        )
        input_names = []
        min_max_names = []
        enter_input_name = self.node_name_mapping[original_node].node.input[1]
        if enter_input_name[0] != "^":
            if self.node_name_mapping[original_node].node.op == "MatMul":
                # mkl ops _MklQuantizedMatMulWithBiasAndRelu|AndRequantize
                # requires the T1 data type as quint8
                dtype = dtypes.quint8
            else:
                input_node_name = helper.node_name_from_input(enter_input_name)
                if input_node_name in self.output_node_maps:
                    if self.output_node_maps[input_node_name].op == "Dequantize":
                        dtype = dtypes.DType(self.output_node_maps[input_node_name].attr["T"].type)
                    elif self._find_relu_node(self.node_name_mapping[original_node].node):
                        dtype = dtypes.quint8
                    else:
                        dtype = dtypes.qint8
                else:
                    dtype = (
                        dtypes.quint8
                        if self._find_relu_node(self.node_name_mapping[original_node].node)
                        else dtypes.qint8
                    )
                    if "FusedBatchNorm" in self.node_name_mapping[original_node].node.op:
                        dtype = dtypes.qint8
            quantize_input_name, min_input_name, max_input_name = self._eightbitize_input_to_node(
                namespace_prefix, enter_input_name, reshape_dims_name, reduction_dims_name, dtype=dtype
            )
            input_names.append(quantize_input_name)
            min_max_names.append(min_input_name)
            min_max_names.append(max_input_name)

        all_input_names = []
        all_input_names.extend(input_names)
        if min_max_names:
            all_input_names.extend(min_max_names)

        for original_input_name in self.node_name_mapping[original_node].node.input:
            if original_input_name[0] == "^":
                all_input_names.append(original_input_name)
        return all_input_names

    def _add_common_quantization_nodes(self, namespace_prefix, control_input_names=None, enter_node=None):
        """Builds constant nodes needed for quantization of inputs."""
        reshape_dims_name = namespace_prefix + "_reshape_dims"
        reduction_dims_name = namespace_prefix + "_reduction_dims"

        reshape_dims_node = helper.create_constant_node(reshape_dims_name, -1, dtypes.int32, [1])

        if enter_node:  # pragma: no cover
            reshape_dims_enter_node = helper.create_node("Enter", reshape_dims_name + "_enter", [reshape_dims_name])
            helper.set_attr_string(reshape_dims_enter_node, "frame_name", enter_node.attr["frame_name"].s)
            helper.set_attr_dtype(reshape_dims_enter_node, "T", dtypes.int32)
            helper.set_attr_bool(reshape_dims_enter_node, "is_constant", True)
            helper.set_attr_int(
                reshape_dims_enter_node, "parallel_iterations", enter_node.attr["parallel_iterations"].i
            )
            self.add_output_graph_node(reshape_dims_enter_node)

        self.add_output_graph_node(reshape_dims_node)
        reduction_dims_node = helper.create_constant_node(reduction_dims_name, 0, dtypes.int32, [1])

        if enter_node:  # pragma: no cover
            reduction_dims_enter_node = helper.create_node(
                "Enter", reduction_dims_name + "_enter", [reduction_dims_name]
            )
            helper.set_attr_string(reduction_dims_enter_node, "frame_name", enter_node.attr["frame_name"].s)
            helper.set_attr_dtype(reduction_dims_enter_node, "T", dtypes.int32)
            helper.set_attr_bool(reduction_dims_enter_node, "is_constant", True)
            helper.set_attr_int(
                reduction_dims_enter_node, "parallel_iterations", enter_node.attr["parallel_iterations"].i
            )
            self.add_output_graph_node(reduction_dims_enter_node)

        self.add_output_graph_node(reduction_dims_node)
        return (
            reshape_dims_enter_node.name if enter_node else reshape_dims_name,
            reduction_dims_enter_node.name if enter_node else reduction_dims_name,
        )

    def add_output_graph_node(self, output_node):
        """Inserts one node into the new graph."""
        assert output_node.name not in self.output_node_maps
        self.output_node_maps[output_node.name] = output_node
        self.output_graph.node.extend([output_node])

    def _parse_graph(self, input_graph=None):
        """Parse the graph and get the input node and output node name details."""
        self.logger.debug("Start to parse graph.")

        graph = self.input_graph if input_graph is None else input_graph
        self.node_name_mapping = {}

        for node in graph.node:
            # each_node = self.node_details(node=node,  output=[])
            if node.name in self.node_name_mapping:
                raise ValueError("Duplicate Node Found when _parse_graph, the node name is {}".format(node.name))
            self.node_name_mapping[node.name] = self.node_details(node=node, output=[])
        for node_name in self.node_name_mapping:
            for each_input in self.node_name_mapping[node_name].node.input:
                self.node_name_mapping[helper.node_name_from_input(each_input)].output.append(node_name)

    def remove_redundant_quantization(self, old_graph):
        """Remove the redundant Quantize/QuantizeV2 nodes."""
        old_nodes_map = self.create_nodes_map(old_graph)
        self.output_graph = graph_pb2.GraphDef()
        inputs_to_rename = {}
        # We go through all the nodes, looking for any that match the patterns we
        # know how to optimize away.
        for node in old_graph.node:
            # We always start with a Quantize node, and examine its inputs to see if
            # they are in a form that can be removed.
            if node.op not in ["Quantize", "QuantizeV2"]:
                continue

            dequantize_node_name = helper.node_name_from_input(node.input[0])

            assert dequantize_node_name in old_nodes_map, (
                "Input node name '" + dequantize_node_name + "' not found in node '" + node.name + "'"
            )

            dequantize_node = old_nodes_map[dequantize_node_name]
            # Do we have a Dequantize feeding in, with the same type as the
            # Quantize?
            if dequantize_node.op != "Dequantize":
                continue

            if node.attr["T"] != dequantize_node.attr["T"]:
                continue

            # Now look at the other inputs, and ensure they're Min/Max nodes.
            min_node_name = helper.node_name_from_input(node.input[1])
            max_node_name = helper.node_name_from_input(node.input[2])
            min_node = old_nodes_map[min_node_name]
            max_node = old_nodes_map[max_node_name]
            is_min_right_type = min_node.op in ["Min", "Dequantize"]
            is_max_right_type = max_node.op in ["Max", "Dequantize"]
            if not is_min_right_type or not is_max_right_type:
                self.logger.info("Not find expected types on inputs {}, {}.".format(min_node.op, max_node.op))
                continue
            min_node_input_name = helper.node_name_from_input(min_node.input[0])
            max_node_input_name = helper.node_name_from_input(max_node.input[0])
            # There are two different patterns for Min nodes we can recognize, one
            # where the input comes directly from the same one as the Max, and
            # another where we run it through another Min first, so check for
            # both.
            is_same_input = False
            if min_node_input_name == max_node_input_name:
                is_same_input = True
            else:
                first_min_node_input = old_nodes_map[min_node_input_name]
                if first_min_node_input.op == "Concat":
                    second_min_node_name = helper.node_name_from_input(first_min_node_input.input[1])
                    second_min_node = old_nodes_map[second_min_node_name]
                    if second_min_node.op == "Min":
                        second_min_node_input_name = helper.node_name_from_input(second_min_node.input[0])
                        is_same_input = second_min_node_input_name == max_node_input_name
            if not is_same_input:
                self.logger.info("Different min/max inputs {}.".format(min_node_input_name))
                continue
            # We recognize this pattern, so mark the graph edges to be rewired to
            # route around it entirely, since we know it's a no-op.
            dequantize_source_name = helper.node_name_from_input(dequantize_node.input[0])
            node_tensor_name = helper.ensure_tensor_name_has_port(node.name)
            min_tensor_name = node.name + ":1"
            max_tensor_name = node.name + ":2"

            inputs_to_rename[node_tensor_name] = dequantize_source_name
            inputs_to_rename[min_tensor_name] = dequantize_node.input[1]
            inputs_to_rename[max_tensor_name] = dequantize_node.input[2]
        # Finally we apply all the rewiring we've marked to the graph.
        for node in old_graph.node:
            for index, input_full_name in enumerate(node.input):
                input_name = helper.ensure_tensor_name_has_port(input_full_name)
                if input_name in inputs_to_rename:
                    node.input[index] = inputs_to_rename[input_name]
            self.add_output_graph_node(node)
        return self.output_graph

    def create_nodes_map(self, graph):
        """Builds a mapping of node names to their defs from the graph."""
        nodes_map = {}
        for node in graph.node:
            assert node.name not in nodes_map, "Duplicate node names detected."
            nodes_map[node.name] = node

        return nodes_map

    def _add_quantize_down_nodes(
        self, original_node, quantized_output_name, requantize_type=dtypes.quint8, is_relu6=False
    ):
        """Add quantize down nodes."""
        quantized_outputs = [quantized_output_name, quantized_output_name + ":1", quantized_output_name + ":2"]
        if not self.fake_quant:
            # Add a RequantizationRange node for finding the min and max values.
            requant_range_node = helper.create_node(
                (
                    "RequantizationRangePerChannel"
                    if self.per_channel or original_node.op == "DepthwiseConv2dNative"
                    else "RequantizationRange"
                ),
                original_node.name + "_eightbit_requant_range",
                quantized_outputs,
            )

            if self.per_channel or original_node.op == "DepthwiseConv2dNative":
                helper.set_attr_dtype(requant_range_node, "T", dtypes.qint32)
                if is_relu6:
                    helper.set_attr_float(requant_range_node, "clip_value_max", 6.0)
                else:
                    helper.set_attr_float(requant_range_node, "clip_value_max", 1e30)
            else:
                helper.set_attr_dtype(requant_range_node, "Tinput", dtypes.qint32)

            self.add_output_graph_node(requant_range_node)
            min_max_inputs = [requant_range_node.name + ":0", requant_range_node.name + ":1"]
        else:
            max_input_name = original_node.name + "_max"
            max_node = helper.create_constant_node(max_input_name, 1.0, dtypes.float32)
            self.add_output_graph_node(max_node)

            min_input_name = original_node.name + "_min"
            min_node = helper.create_constant_node(min_input_name, -1.0, dtypes.float32)
            self.add_output_graph_node(min_node)

            min_max_inputs = [min_input_name, max_input_name]

        requantize_node = helper.create_node(
            "RequantizePerChannel" if self.per_channel or original_node.op == "DepthwiseConv2dNative" else "Requantize",
            original_node.name + "_eightbit_requantize",
            quantized_outputs + min_max_inputs,
        )
        if self.per_channel or original_node.op == "DepthwiseConv2dNative":
            helper.set_attr_dtype(requantize_node, "T", dtypes.qint32)
        else:
            helper.set_attr_dtype(requantize_node, "Tinput", dtypes.qint32)

        helper.set_attr_dtype(requantize_node, "out_type", requantize_type)
        self.add_output_graph_node(requantize_node)
        return requantize_node.name

    def _eightbitize_input_to_node(
        self, namespace_prefix, original_input_name, reshape_dims_name, reduction_dims_name, dtype=dtypes.quint8
    ):
        """Takes one float input to an op, and converts it to quantized form."""
        unique_input_name = helper.unique_node_name_from_input(original_input_name)
        if unique_input_name in self.quantized_node_dict:
            quantized_tuple = self.quantized_node_dict[unique_input_name]
            return quantized_tuple[0], quantized_tuple[1], quantized_tuple[2]

        if self.fake_quant:
            min_input_name = namespace_prefix + "_min_" + unique_input_name
            min_node = helper.create_constant_node(min_input_name, -1.0, dtypes.float32)
            self.add_output_graph_node(min_node)
            max_input_name = namespace_prefix + "_max_" + unique_input_name
            max_node = helper.create_constant_node(max_input_name, 1.0, dtypes.float32)
            self.add_output_graph_node(max_node)
            quantize_input_name = namespace_prefix + "_quantize_" + unique_input_name
            quantize_input_node = helper.create_node(
                "QuantizeV2", quantize_input_name, [original_input_name, min_input_name, max_input_name]
            )
        else:
            reshape_input_name = namespace_prefix + "_reshape_" + unique_input_name
            min_input_name = namespace_prefix + "_min_" + unique_input_name
            max_input_name = namespace_prefix + "_max_" + unique_input_name
            quantize_input_name = namespace_prefix + "_quantize_" + unique_input_name
            reshape_input_node = helper.create_node(
                "Reshape", reshape_input_name, [original_input_name, reshape_dims_name]
            )
            helper.set_attr_dtype(reshape_input_node, "T", dtypes.float32)
            self.add_output_graph_node(reshape_input_node)
            min_input_node = helper.create_node("Min", min_input_name, [reshape_input_name, reduction_dims_name])
            helper.set_attr_dtype(min_input_node, "T", dtypes.float32)
            helper.set_attr_dtype(min_input_node, "Tidx", dtypes.int32)
            helper.set_attr_bool(min_input_node, "keep_dims", False)
            self.add_output_graph_node(min_input_node)
            max_input_node = helper.create_node("Max", max_input_name, [reshape_input_name, reduction_dims_name])
            helper.set_attr_dtype(max_input_node, "T", dtypes.float32)
            helper.set_attr_dtype(max_input_node, "Tidx", dtypes.int32)
            helper.set_attr_bool(max_input_node, "keep_dims", False)
            self.add_output_graph_node(max_input_node)
            quantize_input_node = helper.create_node(
                "QuantizeV2", quantize_input_name, [original_input_name, min_input_name, max_input_name]
            )

        helper.set_attr_dtype(quantize_input_node, "T", dtype)
        helper.set_attr_string(quantize_input_node, "mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED")
        if not self.is_asymmetric:
            helper.set_attr_string(quantize_input_node, "round_mode", b"HALF_TO_EVEN")
        # if FLAGS.model_name in ["wide_deep_large_ds"]:
        #    set_attr_string(quantize_input_node, "mode", b"MIN_FIRST")
        # else:
        #    set_attr_string(quantize_input_node, "mode",
        #                    b"SCALED" if self.intel_cpu_eightbitize else b"MIN_FIRST")
        #    set_attr_string(quantize_input_node, "round_mode",
        #                    b"HALF_TO_EVEN" if self.intel_cpu_eightbitize
        #                    else b"HALF_AWAY_FROM_ZERO")
        self.add_output_graph_node(quantize_input_node)
        min_output_name = quantize_input_name + ":1"
        max_output_name = quantize_input_name + ":2"
        self.quantized_node_dict[unique_input_name] = (quantize_input_name, min_output_name, max_output_name)
        return quantize_input_name, min_output_name, max_output_name

    def _intel_cpu_quantize_weight_eightbit(self, parent, input_node, per_channel, enter_node=None):
        """Quantize weight node."""
        (
            qint8_const_node,
            min_node,
            max_node,
            qint8_const_enter_node,
            min_enter_node,
            max_enter_node,
        ) = helper.generate_quantized_weight_node(
            parent, input_node, per_channel, self.weight_bit, self.device, enter_node
        )
        if enter_node:  # pragma: no cover
            self.add_output_graph_node(qint8_const_node)
            self.add_output_graph_node(qint8_const_enter_node)
            self.add_output_graph_node(min_node)
            self.add_output_graph_node(min_enter_node)
            self.add_output_graph_node(max_node)
            self.add_output_graph_node(max_enter_node)

            return qint8_const_enter_node.name, min_enter_node.name, max_enter_node.name
        else:
            self.add_output_graph_node(qint8_const_node)
            self.add_output_graph_node(min_node)
            self.add_output_graph_node(max_node)

            return qint8_const_node.name, min_node.name, max_node.name



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pre Optimization Entrance."""

import copy
import logging

import tensorflow as tf

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.utils import (
    dump_elapsed_time,
    version1_eq_version2,
    version1_gte_version2,
    version1_lt_version2,
)

from .convert_add_to_biasadd import ConvertAddToBiasAddOptimizer
from .convert_layout import ConvertLayoutOptimizer
from .convert_leakyrelu import ConvertLeakyReluOptimizer
from .convert_nan_to_random import ConvertNanToRandom
from .convert_placeholder_to_const import ConvertPlaceholderToConst
from .dilated_contraction import DilatedContraction
from .dummy_biasadd import InjectDummyBiasAddOptimizer
from .expanddims_optimizer import ExpandDimsOptimizer
from .fetch_weight_from_reshape import FetchWeightFromReshapeOptimizer
from .fold_batch_norm import FoldBatchNormNodesOptimizer
from .fold_constant import GraphFoldConstantOptimizer
from .fuse_biasadd_add import FuseBiasAddAndAddOptimizer
from .fuse_column_wise_mul import FuseColumnWiseMulOptimizer
from .fuse_conv_with_math import FuseConvWithMathOptimizer
from .fuse_decomposed_bn import FuseDecomposedBNOptimizer
from .fuse_decomposed_in import FuseDecomposedINOptimizer
from .fuse_gelu import FuseGeluOptimizer
from .fuse_layer_norm import FuseLayerNormOptimizer
from .fuse_reshape_transpose import FuseTransposeReshapeOptimizer
from .graph_cse_optimizer import GraphCseOptimizer
from .grappler_pass import GrapplerOptimizer
from .move_squeeze_after_relu import MoveSqueezeAfterReluOptimizer
from .remove_training_nodes import RemoveTrainingNodesOptimizer
from .rename_batch_norm import RenameBatchNormOptimizer
from .split_shared_input import SplitSharedInputOptimizer
from .strip_equivalent_nodes import StripEquivalentNodesOptimizer
from .strip_unused_nodes import StripUnusedNodesOptimizer
from .switch_optimizer import SwitchOptimizer


class PreOptimization:
    """Pre optimization for the FP32 models."""

    def __init__(self, model, new_api, device):
        """Initialization."""
        self.model = model
        if version1_gte_version2(tf.version.VERSION, "2.1.0") or version1_eq_version2(tf.version.VERSION, "1.15.0-up3"):
            self.optimization = {
                "pruning": True,
                "shape": True,
                "constfold": False,
                "arithmetic": False,
                "dependency": True,
                "debug_stripper": True,
                "loop": True,
            }
        else:  # pragma: no cover
            self.optimization = {
                "pruning": True,
                "shape": True,
                "dependency": True,
                "debug_stripper": True,
                "loop": True,
            }
        # Table initialization should disable grappler dependency and pruning pass
        node_names = [node.name for node in model.graph_def.node]
        if "init_all_tables" in node_names:  # pragma: no cover
            self.optimization["dependency"] = False
            self.optimization["pruning"] = False
        self.new_api = new_api
        self.device = device
        self.analyzer = GraphAnalyzer()
        self.analyzer.graph = model.graph_def
        self.analyzer.parse_graph()
        self._tmp_graph_def = None
        self._excluded_node_names = []

    def get_excluded_node_names(self):
        """Get the excluded node name.

        Returns:
            string list: the excluded ops' name
        """
        return self._excluded_node_names

    @dump_elapsed_time("Pass Pre Optimization")
    def get_optimized_model(self, itex_mode=False):
        """Executed the non-precision dependent graph optimization.

        The input graph will be optimized with following passes:
        1. Remove the training nodes like Identity Op.
        2. Split the shared nodes like weights node for multi-Conv2d.
        3. Fold Constant Nodes as less as possible.
        4. Fuse the Mul node into the previous Conv2D/MatMul if possible.
        5. Strip the useless nodes.
        6. Do the Common sequence elimation optimization on the graph.
        7. Fold the BN node into the previous Conv2D if possible.

        Returns:
            [graphdef]: the optimized graphdef object.
        """
        from neural_compressor.tensorflow.utils import Model

        origin_model = Model(self.model._model, **self.model.kwargs, backend="itex" if itex_mode else "default")
        origin_model.name = self.model.name
        origin_model.model_type = self.model.model_type
        origin_model.output_tensor_names = self.model.output_tensor_names
        origin_model.input_tensor_names = self.model.input_tensor_names
        origin_model.workspace_path = self.model.workspace_path

        output_node_names = self.model.output_node_names
        input_node_names = self.model.input_node_names
        input_output_names = output_node_names + input_node_names

        # Add device info before convert layout
        # Google in layout optimizer where all nodes in the graph are expected to have their device
        # information set (earlier version < 2.10.0 this was not needed).
        if version1_gte_version2(tf.version.VERSION, "2.10.0"):
            cur_graph = GraphAnalyzer()
            cur_graph.graph = self.model.graph_def
            graph_info = cur_graph.parse_graph()

            if self.device == "cpu":
                cpus = tf.config.list_physical_devices("CPU")
                node_device = cpus[0].name.replace("physical_device:", "")
            else:  # pragma: no cover
                gpus = tf.config.list_physical_devices("GPU")
                if len(gpus) == 0:
                    xpus = tf.config.list_physical_devices("XPU")
                    if len(xpus) == 0:
                        cpus = tf.config.list_physical_devices("CPU")
                        node_device = cpus[0].name.replace("physical_device:", "")
                    else:
                        node_device = xpus[0].name.replace("physical_device:", "")
                else:
                    node_device = gpus[0].name.replace("physical_device:", "")
            for node_name in list(graph_info.keys()):
                node = graph_info[node_name].node
                node.device = node_device
            self._tmp_graph_def = cur_graph.dump_graph()

            self._tmp_graph_def = ConvertLayoutOptimizer(self._tmp_graph_def, output_node_names).do_transformation()
        else:
            self._tmp_graph_def = ConvertLayoutOptimizer(self.model.graph_def, output_node_names).do_transformation()

        self._tmp_graph_def = ConvertPlaceholderToConst(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = SwitchOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = GrapplerOptimizer(
            self._tmp_graph_def, input_output_names, self.optimization
        ).do_transformation()

        self._tmp_graph_def = StripUnusedNodesOptimizer(
            self._tmp_graph_def, input_node_names, output_node_names
        ).do_transformation()

        self._tmp_graph_def = RemoveTrainingNodesOptimizer(
            self._tmp_graph_def, protected_nodes=input_output_names
        ).do_transformation()

        self._tmp_graph_def = SplitSharedInputOptimizer(self._tmp_graph_def).do_transformation()

        # Put FuseDecomposedBNOptimizer before GraphFoldConstantOptimizer
        # The 'Sub' op in the small decomposed ops of BN will be converted to const by GraphFoldConstantOptimizer.
        # Then the FuseDecomposedBNOptimizer can't fuse the small decomposed ops to BN.
        if self.new_api:
            self._tmp_graph_def = FuseDecomposedBNOptimizer(self._tmp_graph_def).do_transformation()
            self._tmp_graph_def = FuseDecomposedINOptimizer(self._tmp_graph_def).do_transformation()
            self._tmp_graph_def = FuseLayerNormOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = GraphFoldConstantOptimizer(self._tmp_graph_def).do_transformation()

        if not self.new_api:
            self._tmp_graph_def = FuseDecomposedBNOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseColumnWiseMulOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = StripUnusedNodesOptimizer(
            self._tmp_graph_def, input_node_names, output_node_names
        ).do_transformation()

        self._tmp_graph_def = FuseGeluOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = GraphCseOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FoldBatchNormNodesOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = RenameBatchNormOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ConvertLeakyReluOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ConvertAddToBiasAddOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseTransposeReshapeOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FuseConvWithMathOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ExpandDimsOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = FetchWeightFromReshapeOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = MoveSqueezeAfterReluOptimizer(self._tmp_graph_def).do_transformation()

        if not self.new_api and not itex_mode:
            # TODO we need to remove below optimizer once the TF enabled the single
            # matmul op quantization
            self._tmp_graph_def = InjectDummyBiasAddOptimizer(
                self._tmp_graph_def, output_node_names
            ).do_transformation()

        self._tmp_graph_def = FuseBiasAddAndAddOptimizer(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = ConvertNanToRandom(self._tmp_graph_def).do_transformation()

        self._tmp_graph_def = StripEquivalentNodesOptimizer(self._tmp_graph_def, output_node_names).do_transformation()

        if self.new_api or itex_mode:
            self._tmp_graph_def = DilatedContraction(self._tmp_graph_def).do_transformation()

        # node device info will be removed by GrapplerOptimizer, insert it again.
        if version1_lt_version2(tf.version.VERSION, "2.0.0"):  # pragma: no cover
            from tensorflow._api.v1.config import experimental

            list_physical_devices = experimental.list_physical_devices
        else:
            list_physical_devices = tf.config.list_physical_devices
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self._tmp_graph_def
        graph_info = cur_graph.parse_graph()

        if self.device == "cpu":
            cpus = list_physical_devices("CPU")
            node_device = cpus[0].name.replace("physical_device:", "")
        else:  # pragma: no cover
            gpus = list_physical_devices("GPU")
            if len(gpus) == 0:
                xpus = list_physical_devices("XPU")
                if len(xpus) == 0:
                    cpus = list_physical_devices("CPU")
                    node_device = cpus[0].name.replace("physical_device:", "")
                else:
                    node_device = xpus[0].name.replace("physical_device:", "")
            else:
                node_device = gpus[0].name.replace("physical_device:", "")
        for node_name in list(graph_info.keys()):
            node = graph_info[node_name].node
            node.device = node_device
        self._tmp_graph_def = cur_graph.dump_graph()

        self._tmp_graph_def.library.CopyFrom(self.model.graph_def.library)

        for function_def in self.model.graph_def.library.function:
            if function_def.signature.name == "swish_f32":  # pragma: no cover
                self._tmp_graph_def.library.function.extend([copy.deepcopy(function_def)])

        origin_model.graph_def = self._tmp_graph_def
        return origin_model

    def get_matched_nodes(self, patterns):
        """Search the matched nodes with the specified patterns.

        Args:
            patterns ([string list]): The patterns should be illustrated as below.
                [['MatMul'], ("BiasAdd"), ("Relu",)]

        Returns:
            [string list]: It will return the list that contains the matched nodes name
                and pattern. ['matched_node_a_name', 'matched_node_a_name',['MatMul','BiasAdd']]
        """
        self.analyzer.graph = self._tmp_graph_def
        self.analyzer.parse_graph()
        res = []

        for sub_pattern in patterns:
            res.extend([i for i in self.analyzer.query_fusion_pattern_nodes(sub_pattern) if i not in res])
        return res

    def has_positive_input(self, node_name):
        """Check the specified node has the positive input or not.

        Args:
            node_name ([string]): node's name

        Returns:
            [bool]: True if the node has the positive input data,
                    False if the node has the negative input data.
        """
        return self.analyzer.has_positive_input(node_name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convert NAN to random Graph Rewriter."""

import numpy as np
from tensorflow.core.framework import attr_value_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer

from ..graph_base import GraphRewriterBase


class ConvertNanToRandom(GraphRewriterBase):
    """Convert Const node which value consists of NAN to random data."""

    def do_transformation(self):
        """Execute convert NAN to random."""
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()

        target_nodes = cur_graph.query_fusion_pattern_nodes([["Const"]])

        for i in target_nodes:
            const_node = graph_info[i[0]].node
            const_content = tensor_util.MakeNdarray(const_node.attr["value"].tensor)
            if const_content.dtype == np.float32 and np.any(np.isnan(const_content)):
                const_node.attr["value"].CopyFrom(
                    attr_value_pb2.AttrValue(
                        tensor=tensor_util.make_tensor_proto(
                            np.random.rand(*const_content.shape), dtypes.float32, const_content.shape
                        )
                    )
                )

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ExpandDims Graph Rewriter."""


import numpy as np
from tensorflow.python.framework import dtypes

from neural_compressor.tensorflow.quantization.utils.graph_util import GraphAnalyzer
from neural_compressor.tensorflow.quantization.utils.graph_util import GraphRewriterHelper as Helper
from neural_compressor.tensorflow.utils import dump_elapsed_time

from ..graph_base import GraphRewriterBase


class ExpandDimsOptimizer(GraphRewriterBase):
    """Calculate ExpandDims and remove it if its input is weight and next node is Conv2D."""

    @dump_elapsed_time("Pass ExpandDimsOptimizer")
    def do_transformation(self):
        """Handle all ExpandDims ops whose input is weight and output is Conv2D.

        Args:
          input_graph_def (graphdef): graphdef object

        Returns:
           [graphdef]: optimized graph
        """
        cur_graph = GraphAnalyzer()
        cur_graph.graph = self.model

        graph_info = cur_graph.parse_graph()
        target_nodes = cur_graph.query_fusion_pattern_nodes([["ExpandDims"]])

        for node_combination in target_nodes:
            expanddims_node = graph_info[node_combination[0]].node
            dims_node = graph_info[expanddims_node.input[1]].node
            next_node = graph_info[graph_info[node_combination[0]].outputs[0]].node
            # to solve the case that input 0 of ExpandDims is a tensor, not a node
            if expanddims_node.input[0] in graph_info:
                weight_node = graph_info[expanddims_node.input[0]].node
            else:
                continue

            if weight_node.op == "Const" and next_node.op == "Conv2D":
                dims = Helper.values_from_const(dims_node)
                weight_value = np.array(Helper.values_from_const(weight_node))
                new_weight_value = np.expand_dims(weight_value, axis=dims)
                cur_graph.remove_node(weight_node.name)
                new_weight_node = Helper.create_constant_node(weight_node.name, new_weight_value, dtypes.float32)

                for output in graph_info[node_combination[0]].outputs:
                    successor_node = graph_info[output].node
                    replace_index = None
                    for index, value in enumerate(successor_node.input):
                        if value == expanddims_node.name:
                            replace_index = index
                            break
                    # weight->conv2d
                    cur_graph.add_node(new_weight_node, None, [successor_node.name])
                    successor_node.input[replace_index] = new_weight_node.name
                # remove ExpandDims and weight_node
                cur_graph.remove_node(expanddims_node.name)
            else:
                continue

        return cur_graph.dump_graph()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantize MatMul/BatchMatMul/BatchMatMulV2."""

import numpy as np
from tensorflow.core.framework import graph_pb2, node_def_pb2
from tensorflow.python.framework import dtypes, tensor_util

from neural_compressor.tensorflow.quantization.utils.quantize_graph_common import QuantizeGraphHelper as helper

from ..quantize_graph_base import QuantizeNodeBase


class FuseNodeStartWithMatmul(QuantizeNodeBase):
    """Quantize MatMul/BatchMatMul/BatchMatMulV2 and apply the fusion."""

    exclude_matmul_nodes = []

    def __init__(self, **kwargs):
        """Initialization."""
        super().__init__(**kwargs)

        self.sorted_patterns = sorted(self.patterns, key=lambda i: len(i), reverse=True)
        self.fusion_op_type = set(fusion[1] for fusion in self.patterns)

        self.fusion_mapping = {
            "DequantizeMatMulBiasAddQuantizeV2": self.apply_matmul_biasadd_fusion,
            "DequantizeMatMulQuantizeV2": self.apply_matmul_biasadd_fusion,
            "DequantizeMatMulBiasAddAddQuantizeV2": self.apply_matmul_biasadd_fusion,
            "DequantizeMatMulAddQuantizeV2": self.apply_matmul_biasadd_fusion,
            "DequantizeMatMulBiasAddAddV2QuantizeV2": self.apply_matmul_biasadd_fusion,
            "DequantizeMatMulAddV2QuantizeV2": self.apply_matmul_biasadd_fusion,
            "DequantizeMatMulBiasAddReluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulBiasAddRelu6QuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulBiasAddLeakyReluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulBiasAddGeluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulBiasAddEluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulBiasAddTanhQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulBiasAddSigmoidQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulReluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulRelu6QuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulLeakyReluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulGeluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulEluQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulTanhQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeMatMulSigmoidQuantizeV2": self.apply_matmul_biasadd_relu_fusion,
            "DequantizeBatchMatMulQuantizeV2": self.apply_batchmatmulv2_fusion,
            "DequantizeBatchMatMulV2QuantizeV2": self.apply_batchmatmulv2_fusion,
            "DequantizeBatchMatMulMulQuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulV2MulQuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulAddQuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulV2AddQuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulAddV2QuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulV2AddV2QuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulMulAddV2QuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulV2MulAddV2QuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulMulAddQuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
            "DequantizeBatchMatMulV2MulAddQuantizeV2": self.apply_batchmatmulv2_mul_add_fusion,
        }

    def apply_matmul_biasadd_relu_fusion(self, match_node_name):
        """Apply dequantize + matmul + biasadd + activation + quantizev2 fusion.

        Dequantize + MatMul + BiasAdd + Relu + QuantizeV2
        Dequantize + MatMul + Relu + QuantizeV2
        Dequantize + MatMul + BiasAdd + Relu6 + QuantizeV2
        Dequantize + MatMul + Relu6 + QuantizeV2
        Dequantize + MatMul + BiasAdd + LeakyRelu + QuantizeV2
        Dequantize + MatMul + LeakyRelu + QuantizeV2
        Dequantize + MatMul + BiasAdd + Gelu + QuantizeV2
        Dequantize + MatMul + Gelu + QuantizeV2
        Dequantize + MatMul + BiasAdd + Elu + QuantizeV2
        Dequantize + MatMul + Elu + QuantizeV2
        Dequantize + MatMul + BiasAdd + Tanh + QuantizeV2
        Dequantize + MatMul + Tanh + QuantizeV2
        Dequantize + MatMul + BiasAdd + Sigmoid + QuantizeV2
        Dequantize + MatMul + Sigmoid + QuantizeV2
        """
        matched_node = self.node_name_mapping[match_node_name[1]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)

        # QDQ inserted for input0 in phase 1
        _, q_inputs = self._get_node_input(normal_inputs[0])

        # Three scenarios to quantize input1:
        #   a. weight node is const, insert QDQ directly in phase 1
        #   b. weight node is non-const, insert QDQ directly in phase 1
        #   c. weight node is 'Enter' and its parent node is const,
        #      not insert QDQ in phase 1 for easy handling
        weight_name = normal_inputs[1]
        weight_node = self.node_name_mapping[helper.node_name_from_input(weight_name)].node
        enter_node = None
        quantizev2_weights_name = None
        weights_min_name = None
        weights_max_name = None
        # no QDQ inserted for 'Enter' node in phase 1
        if weight_node.op == "Enter":  # pragma: no cover
            parent_node = self.node_name_mapping[helper.node_name_from_input(weight_node.input[0])].node
            # FIXME We only quantize the MatMul op which second input node type is const. This is a
            # workaround for RNN model like LTSM.
            if parent_node.op != "Const":
                self.logger.debug("The weight node of matched_node {} is not Const or Const + Enter, skipped")
                self.exclude_matmul_nodes.append(matched_node.node.name)
                self.output_graph = self.input_graph
                return []
            enter_node = weight_node
            weight_node = parent_node
            weight_name = weight_node.name
        # QDQ inserted for other weight nodes in phase 1
        else:
            _, q_weights_inputs = self._get_node_input(weight_name)
            quantizev2_weights_name = q_weights_inputs[0]

            _, weights_name = self._get_node_input(quantizev2_weights_name)
            weights_min_name = weights_name[1]
            weights_max_name = weights_name[2]
            weight_node = self.node_name_mapping[helper.node_name_from_input(weights_name[0])].node
            weight_name = weight_node.name

        if weight_node.op == "Const":
            weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)

            if np.any(np.isnan(weights_content)):  # pragma: no cover
                self.exclude_matmul_nodes.append(matched_node.node.name)
                self.output_graph = self.input_graph
                return []

        # If weight node non const, can't insert dummy biasadd to do matmul fusion.
        if weight_node.op != "Const" and len(match_node_name) == 3:
            self.exclude_matmul_nodes.append(matched_node.node.name)
            self.output_graph = self.input_graph
            return []

        second_node = self.node_name_mapping[match_node_name[2]].node
        skip_node_name = match_node_name[2:]

        need_insert_dummy_biasadd = 1
        offset = 1
        if len(match_node_name) == 5:
            add_a_node_name = helper.node_name_from_input(second_node.input[0])
            add_a_node = self.node_name_mapping[add_a_node_name].node
            add_b_node_name = helper.node_name_from_input(second_node.input[1])
            add_b_node = self.node_name_mapping[add_b_node_name].node
            if (add_a_node.op != "Const" and add_b_node.op == "Const") or (
                add_a_node.op != "Const" and add_b_node.op == "Enter"
            ):
                need_insert_dummy_biasadd = 0
                offset = 0
            if need_insert_dummy_biasadd:
                self.apply_matmul_biasadd_fusion(match_node_name[:2] + [match_node_name[-1]])
                return match_node_name[1:2]

        if weight_node.op == "Const":
            q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
                matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel, enter_node
            )
            if weights_min_name:
                skip_node_name.append(weights_min_name)
            if weights_max_name:
                skip_node_name.append(weights_max_name)
            if quantizev2_weights_name:
                skip_node_name.append(quantizev2_weights_name)
            skip_node_name.append(weight_name)
        else:
            q_weights_name = q_weights_inputs[0]
            q_weights_min_name = q_weights_inputs[1]
            q_weights_max_name = q_weights_inputs[2]

        skip_node_name.append(normal_inputs[0])
        if enter_node:  # pragma: no cover
            skip_node_name.append(enter_node.name)
        else:
            skip_node_name.append(normal_inputs[1])

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_mat_mul"
                if need_insert_dummy_biasadd and weight_node.op == "Const":
                    t_b_index = 0 if matched_node.node.attr["transpose_b"].b else 1
                    weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                    bias_size = weights_content.shape[t_b_index]
                    bias_node_name = node.name + "_fake_bias"
                    bias_node = helper.create_constant_node(
                        bias_node_name, [0] * bias_size, dtypes.float32, shape=[bias_size]
                    )

                    if enter_node:  # pragma: no cover
                        bias_enter_node = helper.create_node("Enter", bias_node_name + "_enter", [bias_node_name])
                        helper.set_attr_string(bias_enter_node, "frame_name", enter_node.attr["frame_name"].s)
                        helper.set_attr_dtype(bias_enter_node, "T", dtypes.float32)
                        helper.set_attr_bool(bias_enter_node, "is_constant", True)
                        helper.set_attr_int(
                            bias_enter_node, "parallel_iterations", enter_node.attr["parallel_iterations"].i
                        )

                        self.add_output_graph_node(bias_enter_node)
                        bias_node_name = bias_enter_node.name

                    self.add_output_graph_node(bias_node)
                else:
                    bias_node_name = self.node_name_mapping[match_node_name[2]].node.input[1]
                relu_node_name = match_node_name[3 - offset]

                all_input_names = q_inputs[:1] + [q_weights_name] + q_inputs[1:]
                all_input_names.append(q_weights_min_name)
                all_input_names.append(q_weights_max_name)
                quantized_node_input_names = (
                    all_input_names[:2] + [bias_node_name] + all_input_names[2:] + control_inputs
                )

                quantized_matmul_node = helper.create_node(
                    "_QuantizedMatMul", quantized_node_name, quantized_node_input_names
                )
                helper.copy_attr(quantized_matmul_node, "transpose_a", node.attr["transpose_a"])
                helper.copy_attr(quantized_matmul_node, "transpose_b", node.attr["transpose_b"])
                helper.set_attr_dtype(quantized_matmul_node, "T1", dtypes.quint8)
                helper.set_attr_dtype(quantized_matmul_node, "T2", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "Tout", dtypes.qint32)
                helper.set_attr_string(
                    quantized_matmul_node, "input_quant_mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED"
                )
                helper.set_attr_string(
                    quantized_matmul_node, "output_quant_mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED"
                )
                if self.node_name_mapping[relu_node_name].node.op == "Relu":
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"Relu"])
                elif self.node_name_mapping[relu_node_name].node.op == "Relu6":
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"Relu6"])
                elif self.node_name_mapping[relu_node_name].node.op == "LeakyRelu":
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"LeakyRelu"])
                elif self.node_name_mapping[relu_node_name].node.op == "Gelu":
                    if self.node_name_mapping[relu_node_name].node.attr["approximate"].b:
                        helper.set_attr_string_list(
                            quantized_matmul_node, "fused_ops", [b"BiasAdd", b"GeluApproximate"]
                        )
                    else:
                        helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"GeluExact"])
                elif self.node_name_mapping[relu_node_name].node.op == "Elu":
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"Elu"])
                elif self.node_name_mapping[relu_node_name].node.op == "Tanh":
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"Tanh"])
                elif self.node_name_mapping[relu_node_name].node.op == "Sigmoid":
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"Sigmoid"])
                helper.set_attr_dtype(quantized_matmul_node, "Tbias", dtypes.float32)
                helper.set_attr_dtype(quantized_matmul_node, "U", dtypes.float32)

                helper.set_attr_type_list(
                    quantized_matmul_node,
                    "Thost_inputs",
                    [
                        dtypes.quint8.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                helper.set_attr_type_list(
                    quantized_matmul_node,
                    "Thost_outputs",
                    [dtypes.qint32.as_datatype_enum, dtypes.float32.as_datatype_enum, dtypes.float32.as_datatype_enum],
                )

                self.add_output_graph_node(quantized_matmul_node)

                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, dtypes.quint8, False)
                self._intel_cpu_add_dequantize_result_node(
                    quantize_down_name, relu_node_name, performance_only=self.performance_only
                )
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                matmul_node = self.node_name_mapping[match_node_name[1]].node
                matmul_node_output = self.node_name_mapping[match_node_name[1]].output
                if new_node.name in matmul_node_output:
                    for idx, node_input in enumerate(new_node.input):
                        if helper.node_name_from_input(node_input) == matmul_node.name:
                            new_node.input[idx] = node_input.replace(matmul_node.name, quantized_node_name)
                self.add_output_graph_node(new_node)
        return match_node_name

    def apply_matmul_biasadd_fusion(self, match_node_name):
        """Apply dequantize + matmul + biasadd + quantizev2 fusion.

        Dequantize + MatMul + QuantizeV2
        Dequantize + MatMul + BiasAdd + QuantizeV2
        Dequantize + MatMul + Add + QuantizeV2
        Dequantize + MatMul + BiasAdd + Add + QuantizeV2
        Dequantize + MatMul + AddV2 + QuantizeV2
        Dequantize + MatMul + BiasAdd + AddV2 + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)

        # QDQ inserted for input0 in phase 1
        _, q_inputs = self._get_node_input(normal_inputs[0])

        weight_name = normal_inputs[1]
        weight_node = self.node_name_mapping[helper.node_name_from_input(weight_name)].node
        enter_node = None
        weights_min_name = None
        weights_max_name = None
        quantizev2_weights_name = None

        # no QDQ inserted for 'Enter' node in phase 1
        if weight_node.op == "Enter":  # pragma: no cover
            parent_node = self.node_name_mapping[helper.node_name_from_input(weight_node.input[0])].node
            # FIXME We only quantize the MatMul op which second input node type is const. This is a
            # workaround for RNN model like LTSM.
            if parent_node.op != "Const":
                self.logger.debug("The weight node of matched_node {} is not Const or Const + Enter, skipped")
                self.exclude_matmul_nodes.append(matched_node.node.name)
                self.output_graph = self.input_graph
                return []
            enter_node = weight_node
            weight_node = parent_node
            weight_name = weight_node.name
        # QDQ inserted for other weight nodes in phase 1
        else:
            _, q_weights_inputs = self._get_node_input(weight_name)
            quantizev2_weights_name = q_weights_inputs[0]

            _, weights_name = self._get_node_input(quantizev2_weights_name)
            weights_min_name = weights_name[1]
            weights_max_name = weights_name[2]
            weight_node = self.node_name_mapping[helper.node_name_from_input(weights_name[0])].node
            weight_name = weight_node.name

        # TODO Remove below two lines once the TF enabled the QuantizedMatMul while
        # transpose_a could be set to True.
        if matched_node.node.attr["transpose_a"].b is True:  # pragma: no cover
            self.exclude_matmul_nodes.append(matched_node.node.name)
            self.output_graph = self.input_graph
            return []

        if weight_node.op == "Const":
            weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)

            if np.any(np.isnan(weights_content)):  # pragma: no cover
                self.exclude_matmul_nodes.append(matched_node.node.name)
                self.output_graph = self.input_graph
                return []

        len_output = len(matched_node.output)
        is_shared_output = False
        if len_output == 2:
            if (
                self.node_name_mapping[matched_node.output[0]].node.op == "Reshape"
                or self.node_name_mapping[matched_node.output[1]].node.op == "Reshape"
            ):
                is_shared_output = False
            else:
                is_shared_output = True
        elif len_output > 1:
            is_shared_output = True

        single_matmul_fusion = True
        if len(match_node_name) == 3:
            if is_shared_output:
                self.output_graph = self.input_graph
                self.exclude_matmul_nodes.append(matched_node.node.name)
                return []
        else:
            second_node = self.node_name_mapping[match_node_name[2]].node
            add_a_node_name = helper.node_name_from_input(second_node.input[0])
            add_a_node = self.node_name_mapping[add_a_node_name].node
            add_b_node_name = helper.node_name_from_input(second_node.input[1])
            add_b_node = self.node_name_mapping[add_b_node_name].node
            if (add_a_node.op != "Const" and add_b_node.op == "Const") or (
                add_a_node.op != "Const" and add_b_node.op == "Enter"
            ):
                single_matmul_fusion = False
            else:
                return self.apply_matmul_biasadd_fusion(match_node_name[:2] + [match_node_name[-1]])

        sum_node_name = ""
        if len(match_node_name) == 4:
            if self.node_name_mapping[match_node_name[2]].node.op in ("Add", "AddV2"):
                sum_index = 1 if match_node_name[1] == self.node_name_mapping[match_node_name[2]].node.input[0] else 0
                sum_node_name = self.node_name_mapping[match_node_name[2]].node.input[sum_index]
                deq_node = self.node_name_mapping[sum_node_name].node
                if deq_node.op != "Dequantize" or deq_node.op.find("Quantize") != -1:
                    return self.apply_matmul_biasadd_fusion(match_node_name[:2] + [match_node_name[-1]])
        if len(match_node_name) == 5:
            if self.node_name_mapping[match_node_name[3]].node.op in ("Add", "AddV2"):
                sum_index = 1 if match_node_name[2] == self.node_name_mapping[match_node_name[3]].node.input[0] else 0
                sum_node_name = self.node_name_mapping[match_node_name[3]].node.input[sum_index]
                deq_node = self.node_name_mapping[sum_node_name].node
                if deq_node.op != "Dequantize" or deq_node.op.find("Quantize") != -1:
                    return self.apply_matmul_biasadd_fusion(match_node_name[:3] + [match_node_name[-1]])

        if weight_node.op == "Const":
            q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
                matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel, enter_node
            )
            if weights_min_name:
                skip_node_name.append(weights_min_name)
            if weights_max_name:
                skip_node_name.append(weights_max_name)
            if quantizev2_weights_name:
                skip_node_name.append(quantizev2_weights_name)
            skip_node_name.append(weight_name)
        else:
            q_weights_name = q_weights_inputs[0]
            q_weights_min_name = q_weights_inputs[1]
            q_weights_max_name = q_weights_inputs[2]

        skip_node_name.append(normal_inputs[0])
        if enter_node:
            if len(self.node_name_mapping[helper.node_name_from_input(enter_node.name)].output) == 1:
                skip_node_name.append(enter_node.name)
        else:
            skip_node_name.append(normal_inputs[1])

        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_mat_mul"
                all_input_names = q_inputs[:1] + [q_weights_name] + q_inputs[1:]
                all_input_names.append(q_weights_min_name)
                all_input_names.append(q_weights_max_name)

                if single_matmul_fusion:
                    if sum_node_name:
                        quantized_node_input_names = (
                            all_input_names[:2] + [sum_node_name] + all_input_names[2:] + control_inputs
                        )
                    else:
                        quantized_node_input_names = all_input_names[:2] + all_input_names[2:] + control_inputs
                else:
                    bias_node_name = self.node_name_mapping[match_node_name[2]].node.input[1]
                    if sum_node_name:
                        quantized_node_input_names = (
                            all_input_names[:2]
                            + [bias_node_name]
                            + [sum_node_name]
                            + all_input_names[2:]
                            + control_inputs
                        )
                    else:
                        quantized_node_input_names = (
                            all_input_names[:2] + [bias_node_name] + all_input_names[2:] + control_inputs
                        )

                quantized_matmul_node = helper.create_node(
                    "_QuantizedMatMul", quantized_node_name, quantized_node_input_names
                )
                helper.copy_attr(quantized_matmul_node, "transpose_a", node.attr["transpose_a"])
                helper.copy_attr(quantized_matmul_node, "transpose_b", node.attr["transpose_b"])
                helper.set_attr_dtype(quantized_matmul_node, "T1", dtypes.quint8)
                helper.set_attr_dtype(quantized_matmul_node, "T2", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "Tout", dtypes.qint32)
                helper.set_attr_dtype(quantized_matmul_node, "U", dtypes.float32)
                helper.set_attr_string(
                    quantized_matmul_node, "input_quant_mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED"
                )
                helper.set_attr_string(
                    quantized_matmul_node, "output_quant_mode", b"MIN_FIRST" if self.is_asymmetric else b"SCALED"
                )
                helper.set_attr_dtype(quantized_matmul_node, "Tbias", dtypes.float32)
                if sum_node_name:
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd", b"Add"])
                    helper.set_attr_type_list(
                        quantized_matmul_node,
                        "Thost_inputs",
                        [
                            dtypes.quint8.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                else:
                    if not single_matmul_fusion:
                        helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"BiasAdd"])
                        helper.set_attr_type_list(
                            quantized_matmul_node,
                            "Thost_inputs",
                            [
                                dtypes.quint8.as_datatype_enum,
                                dtypes.qint8.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                    else:
                        helper.set_attr_type_list(
                            quantized_matmul_node,
                            "Thost_inputs",
                            [
                                dtypes.quint8.as_datatype_enum,
                                dtypes.qint8.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                                dtypes.float32.as_datatype_enum,
                            ],
                        )
                helper.set_attr_type_list(
                    quantized_matmul_node,
                    "Thost_outputs",
                    [dtypes.qint32.as_datatype_enum, dtypes.float32.as_datatype_enum, dtypes.float32.as_datatype_enum],
                )

                self.add_output_graph_node(quantized_matmul_node)
                requantize_type = dtypes.qint8

                quantize_down_name = self._add_quantize_down_nodes(node, quantized_node_name, requantize_type, False)
                if sum_node_name:
                    self._intel_cpu_add_dequantize_result_node(
                        quantize_down_name,
                        match_node_name[2] if single_matmul_fusion else match_node_name[3],
                        requantize_type,
                        performance_only=self.performance_only,
                    )
                else:
                    self._intel_cpu_add_dequantize_result_node(
                        quantize_down_name,
                        match_node_name[1] if single_matmul_fusion else match_node_name[2],
                        requantize_type,
                        performance_only=self.performance_only,
                    )
            else:
                new_node = node_def_pb2.NodeDef()
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)
        return match_node_name

    def apply_batchmatmulv2_fusion(self, match_node_name):  # pragma: no cover
        """Apply dequantize + batchmatmul/batchmatmulv2 + quantizev2 fusion.

        Dequantize + BatchMatMulV2 + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]
        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)

        _, q_x_inputs = self._get_node_input(normal_inputs[0])
        quantizev2_input_name = q_x_inputs[0]
        weight_name = normal_inputs[1]
        weight_node = self.node_name_mapping[helper.node_name_from_input(weight_name)].node
        enter_node = None
        weights_min_name = None
        weights_max_name = None
        quantizev2_y_name = None

        if weight_node.op == "Enter":
            parent_node = self.node_name_mapping[helper.node_name_from_input(weight_node.input[0])].node
            # FIXME We only quantize the MatMul op which second input node type is const. This is a
            # workaround for RNN model like LTSM.
            if parent_node.op != "Const":
                self.logger.debug("The weight node of matched_node {} is not Const or Const + Enter, skipped")
                self.exclude_matmul_nodes.append(matched_node.node.name)
                self.output_graph = self.input_graph
                return []
            enter_node = weight_node
            weight_node = parent_node
            weight_name = weight_node.name
        else:
            _, q_y_inputs = self._get_node_input(normal_inputs[1])
            quantizev2_y_name = q_y_inputs[0]
            _, weights_name = self._get_node_input(quantizev2_y_name)
            weights_min_name = weights_name[1]
            weights_max_name = weights_name[2]
            weight_node = self.node_name_mapping[helper.node_name_from_input(weights_name[0])].node
            weight_name = weight_node.name

        if weight_node.op == "Const":
            weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)

            if np.any(np.isnan(weights_content)):
                self.output_graph = self.input_graph
                self.exclude_matmul_nodes.append(matched_node.node.name)
                return []

            for i in self.node_name_mapping:
                if (
                    weight_node.input
                    and not weight_node.input[0].startswith("^")
                    and weight_node.name in self.node_name_mapping[i].output
                ):
                    self.output_graph = self.input_graph
                    self.exclude_matmul_nodes.append(matched_node.node.name)
                    return []

            q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
                matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel, enter_node
            )

            if weights_min_name:
                skip_node_name.append(weights_min_name)
            if weights_max_name:
                skip_node_name.append(weights_max_name)
            if quantizev2_y_name:
                skip_node_name.append(quantizev2_y_name)
            skip_node_name.append(weight_name)
        else:
            q_weights_name = q_y_inputs[0]
            q_weights_min_name = q_y_inputs[1]
            q_weights_max_name = q_y_inputs[2]

        skip_node_name.append(normal_inputs[0])
        if enter_node:
            skip_node_name.append(enter_node.name)
        else:
            skip_node_name.append(normal_inputs[1])

        batchmatmul_next_node = {}
        for _, node in enumerate(self.input_graph.node):
            quantized_node_name = ""
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))

                quantized_node_name = node.name + "_eightbit_quantized_batch_matmul_v2"
                all_input_names = [quantizev2_input_name] + [q_weights_name] + q_x_inputs[1:]
                all_input_names.append(q_weights_min_name)
                all_input_names.append(q_weights_max_name)

                quantized_node_input_names = all_input_names + control_inputs

                quantized_matmul_node = helper.create_node(
                    "_QuantizedBatchMatMul", quantized_node_name, quantized_node_input_names
                )

                helper.copy_attr(quantized_matmul_node, "adj_x", node.attr["adj_x"])
                helper.copy_attr(quantized_matmul_node, "adj_y", node.attr["adj_y"])
                helper.set_attr_dtype(quantized_matmul_node, "T1", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "T2", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "U", dtypes.float32)
                helper.set_attr_dtype(quantized_matmul_node, "Tout", dtypes.float32)
                helper.set_attr_string(quantized_matmul_node, "input_quant_mode", b"SCALED")
                helper.set_attr_string(quantized_matmul_node, "output_quant_mode", b"SCALED")
                helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"Dequantize"])
                helper.set_attr_type_list(
                    quantized_matmul_node,
                    "Thost_inputs",
                    [
                        dtypes.qint8.as_datatype_enum,
                        dtypes.qint8.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                        dtypes.float32.as_datatype_enum,
                    ],
                )
                helper.set_attr_type_list(quantized_matmul_node, "Thost_outputs", [dtypes.float32.as_datatype_enum])

                self.add_output_graph_node(quantized_matmul_node)
                for i in self.node_name_mapping[node.name].output:
                    batchmatmul_next_node[i] = (quantized_node_name, node.name)
            else:
                new_node = node_def_pb2.NodeDef()
                if batchmatmul_next_node.get(node.name):
                    for index, name in enumerate(node.input):
                        if name == batchmatmul_next_node[node.name][1]:
                            node.input[index] = batchmatmul_next_node[node.name][0]
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)
        return match_node_name

    def apply_batchmatmulv2_mul_add_fusion(self, match_node_name):  # pragma: no cover
        """Apply dequantize + batchmatmul/batchmatmulv2 + mul + add + quantizev2 fusion.

        Dequantize + BatchMatMulV2 + Mul + QuantizeV2
        Dequantize + BatchMatMulV2 + Add + QuantizeV2
        Dequantize + BatchMatMulV2 + AddV2 + QuantizeV2
        Dequantize + BatchMatMulV2 + Mul + Add + QuantizeV2
        Dequantize + BatchMatMulV2 + Mul + AddV2 + QuantizeV2
        """
        skip_node_name = match_node_name[2:]
        matched_node = self.node_name_mapping[match_node_name[1]]
        # oneDNN limitation: add tensor ndim must be 4
        if len(match_node_name) == 4 and self.node_name_mapping[match_node_name[2]].node.op in ("Add", "AddV2"):
            add_node_input_name = self.node_name_mapping[match_node_name[2]].node.input[1]
            if add_node_input_name == matched_node.node.name:
                add_node_input_name = self.node_name_mapping[match_node_name[2]].node.input[0]
            add_input_node = self.node_name_mapping[add_node_input_name].node
            if add_input_node.op != "Const":
                return self.apply_batchmatmulv2_fusion(match_node_name[:2] + [match_node_name[-1]])

            shape = tensor_util.MakeNdarray(add_input_node.attr["value"].tensor)
            if shape.ndim != 4:
                return self.apply_batchmatmulv2_fusion(match_node_name[:2] + [match_node_name[-1]])

        control_inputs, normal_inputs = self._get_node_input(matched_node.node.name)

        weight_name = normal_inputs[1]
        weight_node = self.node_name_mapping[helper.node_name_from_input(weight_name)].node
        _, q_x_inputs = self._get_node_input(normal_inputs[0])
        enter_node = None
        weights_min_name = None
        weights_max_name = None
        quantizev2_weights_name = None
        if weight_node.op == "Enter":
            parent_node = self.node_name_mapping[helper.node_name_from_input(weight_node.input[0])].node
            # FIXME We only quantize the MatMul op which second input node type is const. This is a
            # workaround for RNN model like LTSM.
            if parent_node.op != "Const":
                self.logger.debug("The weight node of matched_node {} is not Const or Const + Enter, skipped")
                self.exclude_matmul_nodes.append(matched_node.node.name)
                self.output_graph = self.input_graph
                return []
            enter_node = weight_node
            weight_node = parent_node
            weight_name = weight_node.name
        else:
            _, q_weights_inputs = self._get_node_input(normal_inputs[1])
            quantizev2_weights_name = q_weights_inputs[0]
            _, weights_name = self._get_node_input(quantizev2_weights_name)
            weights_min_name = weights_name[1]
            weights_max_name = weights_name[2]
            weight_node = self.node_name_mapping[helper.node_name_from_input(weights_name[0])].node
            weight_name = weight_node.name

        if weight_node.op == "Const":
            weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)

            if np.any(np.isnan(weights_content)):
                self.exclude_matmul_nodes.append(matched_node.node.name)
                self.output_graph = self.input_graph
                return []

            for i in self.node_name_mapping:
                if (
                    weight_node.input
                    and not weight_node.input[0].startswith("^")
                    and weight_node.name in self.node_name_mapping[i].output
                ):
                    self.output_graph = self.input_graph
                    self.exclude_matmul_nodes.append(matched_node.node.name)
                    return []

            q_weights_name, q_weights_min_name, q_weights_max_name = self._intel_cpu_quantize_weight_eightbit(
                matched_node.node.op, self.node_name_mapping[weight_name].node, self.per_channel, enter_node
            )

            if weights_min_name:
                skip_node_name.append(weights_min_name)
            if weights_max_name:
                skip_node_name.append(weights_max_name)
            if quantizev2_weights_name:
                skip_node_name.append(quantizev2_weights_name)
            skip_node_name.append(weight_name)
        else:
            q_weights_name = q_weights_inputs[0]
            q_weights_min_name = q_weights_inputs[1]
            q_weights_max_name = q_weights_inputs[2]

        skip_node_name.append(normal_inputs[0])
        if enter_node:
            skip_node_name.append(enter_node.name)
        else:
            skip_node_name.append(normal_inputs[1])

        batchmatmul_next_node = {}
        for _, node in enumerate(self.input_graph.node):
            if node.name in skip_node_name:
                pass
            elif node.name == match_node_name[1]:
                self.logger.debug("Matched node {} with input {}.".format(node.name, node.input))
                quantized_node_name = node.name + "_eightbit_quantized_batch_matmul_v2"

                if len(match_node_name) == 4:
                    if self.node_name_mapping[match_node_name[2]].node.op == "Mul":
                        mul_node_name = self.node_name_mapping[match_node_name[2]].node.input[1]
                        all_input_names = q_x_inputs[:1] + [q_weights_name] + [mul_node_name] + q_x_inputs[1:]
                        all_input_names.append(q_weights_min_name)
                        all_input_names.append(q_weights_max_name)
                    else:
                        add_node_name = self.node_name_mapping[match_node_name[2]].node.input[1]
                        all_input_names = q_x_inputs[:1] + [q_weights_name] + [add_node_name] + q_x_inputs[1:]
                        all_input_names.append(q_weights_min_name)
                        all_input_names.append(q_weights_max_name)
                    skip_node_name.append(match_node_name[2])
                else:
                    mul_node_name = self.node_name_mapping[match_node_name[2]].node.input[1]
                    add_node_name = self.node_name_mapping[match_node_name[3]].node.input[1]
                    skip_node_name.append(match_node_name[2])
                    skip_node_name.append(match_node_name[3])
                    all_input_names = (
                        q_x_inputs[:1] + [q_weights_name] + [mul_node_name] + [add_node_name] + q_x_inputs[1:]
                    )
                    all_input_names.append(q_weights_min_name)
                    all_input_names.append(q_weights_max_name)

                quantized_node_input_names = all_input_names + control_inputs

                quantized_matmul_node = helper.create_node(
                    "_QuantizedBatchMatMul", quantized_node_name, quantized_node_input_names
                )

                helper.copy_attr(quantized_matmul_node, "adj_x", node.attr["adj_x"])
                helper.copy_attr(quantized_matmul_node, "adj_y", node.attr["adj_y"])
                helper.set_attr_dtype(quantized_matmul_node, "T1", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "T2", dtypes.qint8)
                helper.set_attr_dtype(quantized_matmul_node, "U", dtypes.float32)
                helper.set_attr_dtype(quantized_matmul_node, "Tout", dtypes.float32)
                helper.set_attr_string(quantized_matmul_node, "input_quant_mode", b"SCALED")
                helper.set_attr_string(quantized_matmul_node, "output_quant_mode", b"SCALED")
                if len(match_node_name) == 4:
                    if self.node_name_mapping[match_node_name[2]].node.op == "Mul":
                        helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"Mul", b"Dequantize"])
                    else:
                        helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"Add", b"Dequantize"])
                    helper.set_attr_type_list(
                        quantized_matmul_node,
                        "Thost_inputs",
                        [
                            dtypes.qint8.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                else:
                    helper.set_attr_string_list(quantized_matmul_node, "fused_ops", [b"Mul", b"Add", b"Dequantize"])
                    helper.set_attr_type_list(
                        quantized_matmul_node,
                        "Thost_inputs",
                        [
                            dtypes.qint8.as_datatype_enum,
                            dtypes.qint8.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                            dtypes.float32.as_datatype_enum,
                        ],
                    )
                helper.set_attr_type_list(quantized_matmul_node, "Thost_outputs", [dtypes.float32.as_datatype_enum])

                self.add_output_graph_node(quantized_matmul_node)
                attr_fused_ops = "".join(
                    x
                    for x in quantized_matmul_node.attr["fused_ops"]
                    .SerializeToString()
                    .decode("UTF-8", "ignore")
                    .strip()
                    if x.isprintable()
                )
                if "MulAdd" in attr_fused_ops:
                    for i in self.node_name_mapping[match_node_name[3]].output:
                        batchmatmul_next_node[i] = (quantized_node_name, match_node_name[3])
                else:
                    for i in self.node_name_mapping[match_node_name[2]].output:
                        batchmatmul_next_node[i] = (quantized_node_name, match_node_name[2])
            else:
                new_node = node_def_pb2.NodeDef()
                if batchmatmul_next_node.get(node.name):
                    for index, name in enumerate(node.input):
                        if name == batchmatmul_next_node[node.name][1]:
                            node.input[index] = batchmatmul_next_node[node.name][0]
                new_node.CopyFrom(node)
                self.add_output_graph_node(new_node)
        return match_node_name

    def get_longest_fuse(self):
        """Get the longest fusion pattern."""
        self._get_op_list()
        matched_rule, matched_node_name = self._is_match_matmul(self.sorted_patterns)
        return matched_rule, matched_node_name

    def apply_the_transform(self):
        """Quantize MatMul/BatchMatMul/BatchMatMulV2 and apply the fusion pattern."""
        self._get_op_list()
        matched_rule, matched_node_name = self._is_match_matmul(self.sorted_patterns, True)
        if matched_node_name:
            _, normal_inputs = self._get_node_input(matched_node_name[1])
        if matched_node_name and self.node_name_mapping[normal_inputs[0]].node.op == matched_node_name[0]:
            self.output_graph = graph_pb2.GraphDef()
            fusion_name = "".join(matched_rule)
            if fusion_name in self.fusion_mapping:
                _ = self.fusion_mapping[fusion_name](matched_node_name)
            else:  # pragma: no cover
                self.logger.debug("Unknown fusion pattern {}.".format(fusion_name))
                if self.remove_redundant_quant_flag:
                    self.input_graph = self.remove_redundant_quantization(self.input_graph)
                return self.input_graph, self.exclude_matmul_nodes

            self.input_graph = self.output_graph
            self._reset_output_node_maps()
            if self.remove_redundant_quant_flag:
                self.output_graph = self.remove_redundant_quantization(self.output_graph)
            return self.output_graph, self.exclude_matmul_nodes

        if self.remove_redundant_quant_flag:
            self.input_graph = self.remove_redundant_quantization(self.input_graph)
        return self.input_graph, self.exclude_matmul_nodes

    def _is_match_matmul(self, patterns, qdq_inserted=False):
        """Detect the rule matched nodes collections.

        Returns:
            [List] -- [the matched rule]
            [String] -- [the list contains the matched node name]
        """
        matched_node_name = []

        for k, v in enumerate(self.op_list):
            if v in set(fusion[1] for fusion in patterns):
                cur_node = self.node_name_mapping[list(self.node_name_mapping.keys())[k]].node
                if cur_node.name != self.start_node_name:
                    continue

                # Disable BatchMatMul quantization temporarily for its bad performance
                # Enable it again when the performance issue is fixed.
                if cur_node.op in ("BatchMatMulV2", "BatchMatMul"):
                    self.exclude_matmul_nodes.append(cur_node.name)
                    continue

                control_inputs, normal_inputs = self._get_node_input(cur_node.name)
                weight_name = normal_inputs[1]
                weight_node = self.node_name_mapping[helper.node_name_from_input(weight_name)].node
                if not qdq_inserted:
                    # FIXME We only quantize the MatMul op which second input node type is const.
                    # This is a workaround for RNN model like LTSM.
                    parent_node = None
                    if cur_node.op == "MatMul" and not self.itex_mode:
                        if control_inputs:  # pragma: no cover
                            self.exclude_matmul_nodes.append(cur_node.name)
                            continue
                        if weight_node.op != "Const":
                            if weight_node.input:
                                parent_node = self.node_name_mapping[
                                    helper.node_name_from_input(weight_node.input[0])
                                ].node
                                if weight_node.op == "Enter":  # pragma: no cover
                                    if len(self.node_name_mapping[helper.node_name_from_input(weight_name)].output) > 1:
                                        self.exclude_matmul_nodes.append(cur_node.name)
                                        continue
                                    if parent_node.op == "Const":
                                        weight_node = parent_node
                                        weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                                        if np.any(np.isnan(weights_content)):
                                            self.exclude_matmul_nodes.append(cur_node.name)
                                            continue
                                    else:
                                        self.exclude_matmul_nodes.append(cur_node.name)
                                        continue
                        else:
                            weights_content = tensor_util.MakeNdarray(weight_node.attr["value"].tensor)
                            if np.any(np.isnan(weights_content)):  # pragma: no cover
                                self.exclude_matmul_nodes.append(cur_node.name)
                                continue

                        # TODO Remove below two lines once the TF enabled the QuantizedMatMul while
                        # transpose_a could be set to True.
                        if cur_node.attr["transpose_a"].b is True:  # pragma: no cover
                            self.exclude_matmul_nodes.append(cur_node.name)
                            continue

                for sub_rule in patterns:
                    if sub_rule[0] != "Dequantize":
                        self.exclude_matmul_nodes.append(cur_node.name)
                        continue
                    if v != sub_rule[1]:
                        self.exclude_matmul_nodes.append(cur_node.name)
                        continue

                    if qdq_inserted:
                        if control_inputs:  # pragma: no cover
                            self.exclude_matmul_nodes.append(cur_node.name)
                            continue
                        if self.node_name_mapping[normal_inputs[0]].node.op != "Dequantize" or self.node_name_mapping[
                            normal_inputs[1]
                        ].node.op not in ("Dequantize", "Enter"):
                            self.exclude_matmul_nodes.append(cur_node.name)
                            continue

                    sub_rule_len = len(sub_rule) - 2
                    self.logger.debug("Try to apply rule: {}".format(sub_rule))

                    cur_node_name = list(self.node_name_mapping.keys())[k]

                    matched_node_name.clear()
                    matched_node_name.append(sub_rule[0])
                    matched_node_name.append(cur_node_name)

                    while sub_rule_len > 1:
                        if not self.node_name_mapping[cur_node_name].output:
                            self.logger.debug("Fail to match {}".format(sub_rule))
                            break

                        next_node_name = self.node_name_mapping[cur_node_name].output[0]

                        next_node_op = self.node_name_mapping[next_node_name].node.op

                        if next_node_op == sub_rule[-sub_rule_len]:
                            matched_node_name.append(next_node_name)
                            sub_rule_len -= 1
                            cur_node_name = next_node_name
                        else:
                            matched_node_name.clear()
                            self.logger.debug("Fail to match {}.".format(sub_rule))
                            break

                    if sub_rule_len == 1:
                        matched_node_name.append(sub_rule[-1])
                        self.logger.debug("Match {} on nodes {}.".format(sub_rule, matched_node_name))
                        return sub_rule, matched_node_name

        return None, None



#
#  -*- coding: utf-8 -*-
#
#  Copyright (c) 2021 Intel Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
"""Insert logging graph transformation."""

from __future__ import absolute_import, division, print_function

from tensorflow.core.framework import attr_value_pb2, node_def_pb2
from tensorflow.python.framework import dtypes

from .graph_transform_base import GraphTransformBase


class InsertLogging(GraphTransformBase):
    """Insert logging graph transformation."""

    op_output_type_mapping = {
        "RequantizationRange": [dtypes.float32.as_datatype_enum, dtypes.float32.as_datatype_enum],
        "RequantizationRangePerChannel": [dtypes.float32.as_datatype_enum, dtypes.float32.as_datatype_enum],
        "QuantizedConv2DWithBiasAndRelu": [dtypes.qint32.as_datatype_enum],
        "QuantizedConv2DWithBiasAndReluAndRequantize": [dtypes.quint8.as_datatype_enum],
        "QuantizedConv2DWithBiasAndRequantize": [dtypes.qint8.as_datatype_enum],
        "QuantizedConv2DWithBiasSignedSumAndReluAndRequantize": [dtypes.qint8.as_datatype_enum],
        "QuantizedConv2DWithBiasSumAndReluAndRequantize": [dtypes.quint8.as_datatype_enum],
        "QuantizedDepthwiseConv2DWithBiasAndReluAndRequantize": [dtypes.quint8.as_datatype_enum],
        "QuantizedConv2DWithBias": [dtypes.qint32.as_datatype_enum],
        "Relu": [dtypes.float32.as_datatype_enum],
        "Relu6": [dtypes.float32.as_datatype_enum],
        "AvgPool": [dtypes.float32.as_datatype_enum],
        "MaxPool": [dtypes.float32.as_datatype_enum],
        "BiasAdd": [dtypes.float32.as_datatype_enum],
        "Max": [dtypes.float32.as_datatype_enum],
        "Min": [dtypes.float32.as_datatype_enum],
    }

    def __init__(
        self,
        input_pb,
        ops=[],
        node_name_list=[],
        show_name=True,
        show_op=False,
        first_n=-1,
        summarize=1024,
        message="",
        dump_fp32=False,
    ):
        """Initialization."""
        super(InsertLogging, self).__init__(input_pb)

        self.parse_input_pb()
        self.ops = ops
        self.node_name_list = node_name_list
        self.show_name = show_name
        self.show_op = show_op
        self.message = message
        self.first_n = first_n
        self.summarize = summarize
        self.output_name_index_mapping = {}
        self.input_rename = {}
        self.dump_fp32 = dump_fp32

    def _get_suffix(self, input_str):
        """Split the node name into two parts.

        Returns:
            Pure string name without suffix
            Index of the node
        """
        splitted_str = input_str.split(":")
        if len(splitted_str) < 2:
            return input_str, 0

        return splitted_str[0], int(splitted_str[-1])

    def _get_output_index_mapping(self):
        """Get the output_node_name and index mapping."""
        for node_name in self.node_mapping:
            for node_input in self.node_mapping[node_name].input:
                node_stripped_name, suffix = self._get_suffix(node_input)

                if node_stripped_name not in self.output_name_index_mapping:
                    self.output_name_index_mapping[node_stripped_name] = set()

                self.output_name_index_mapping[node_stripped_name].add(suffix)

    def _insert_node(self):
        """Insert the Print OP into the graph."""
        for node_name in self.node_mapping:
            if node_name not in self.output_name_index_mapping or (
                not self.dump_fp32 and node_name.find("eightbit") == -1
            ):
                continue

            if self.ops and self.node_mapping[node_name].op in self.ops or node_name in self.node_name_list:
                name_suffix = "__print__"
                print_node = node_def_pb2.NodeDef()
                print_node.op = "Print"
                print_node.name = node_name + name_suffix
                node_message = ""
                if self.show_op:
                    node_message += ";" + self.node_mapping[node_name].op + ";"
                if self.show_name:
                    node_message += ";" + print_node.name + ";"
                node_message += self.message

                print_node.attr["message"].s = node_message.encode()
                print_node.attr["first_n"].i = self.first_n
                print_node.attr["summarize"].i = self.summarize

                print_node.input.append(node_name + ":0")
                print_node.attr["T"].CopyFrom(
                    attr_value_pb2.AttrValue(type=self.op_output_type_mapping[self.node_mapping[node_name].op][0])
                )

                if self.node_mapping[node_name].op in ("QuantizedConv2DWithBias", "QuantizedConv2DWithBiasAndRelu"):
                    for index in sorted(self.output_name_index_mapping[node_name])[:1]:
                        print_node.input.append(node_name + ":" + str(index))

                        print_node_1 = node_def_pb2.NodeDef()
                        print_node_1.op = "Print"
                        print_node_1.name = node_name + name_suffix + "_min_output"

                        print_node_1.attr["message"].s = (node_message + "_min_output").encode()
                        print_node_1.attr["first_n"].i = self.first_n
                        print_node_1.attr["summarize"].i = self.summarize
                        print_node_1.attr["U"].list.CopyFrom(
                            attr_value_pb2.AttrValue.ListValue(type=[dtypes.float32.as_datatype_enum])
                        )
                        print_node_1.attr["T"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
                        print_node_1.input.append(node_name + ":1")
                        print_node_1.input.append(node_name + ":1")
                        self.input_graph.node.extend([print_node_1])
                        self.input_rename[node_name + ":1"] = print_node_1.name + ":0"

                        print_node_2 = node_def_pb2.NodeDef()
                        print_node_2.op = "Print"
                        print_node_2.name = node_name + name_suffix + "_max_output"

                        print_node_2.attr["message"].s = (node_message + "_max_output").encode()
                        print_node_2.attr["first_n"].i = self.first_n
                        print_node_2.attr["summarize"].i = self.summarize
                        print_node_2.attr["U"].list.CopyFrom(
                            attr_value_pb2.AttrValue.ListValue(type=[dtypes.float32.as_datatype_enum])
                        )
                        print_node_2.attr["T"].CopyFrom(attr_value_pb2.AttrValue(type=dtypes.float32.as_datatype_enum))
                        print_node_2.input.append(node_name + ":2")
                        print_node_2.input.append(node_name + ":2")
                        self.input_graph.node.extend([print_node_2])
                        self.input_rename[node_name + ":2"] = print_node_2.name + ":0"
                else:
                    for index in range(len(self.op_output_type_mapping[self.node_mapping[node_name].op])):
                        print_node.input.append(
                            node_name + ":" + str(sorted(self.output_name_index_mapping[node_name])[index])
                        )

                print_node.attr["U"].list.CopyFrom(
                    attr_value_pb2.AttrValue.ListValue(
                        type=self.op_output_type_mapping[self.node_mapping[node_name].op]
                    )
                )

                self.input_graph.node.extend([print_node])

                self.input_rename[node_name + ":0"] = print_node.name + ":0"

    def _rename_node(self):
        """Rename the original input node and connect to new added print node."""
        for node_name in self.node_mapping:
            for index, input_name in enumerate(self.node_mapping[node_name].input):
                if input_name in self.input_rename:
                    self.node_mapping[node_name].input[index] = self.input_rename[input_name]
                elif input_name + ":0" in self.input_rename:
                    self.node_mapping[node_name].input[index] = self.input_rename[input_name + ":0"]

    def do_transformation(self):
        """Execute the insert logging transformation.

        :return: Transformed graph
        """
        self._get_output_index_mapping()
        self._insert_node()
        self._rename_node()

        return self.input_graph
